"""
Class for managing notifications in pegasus-monitord.
"""

##
#  Copyright 2007-2011 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
##

import logging
import math
import os
import shlex
import signal
import subprocess
import sys
import tempfile
import time

from Pegasus.tools import utils

NOTIFICATION_FILE = "monitord-notifications.log"  # filename for writing the output of notification scripts
WAIT_CHILD_FINISH = 5  # in seconds

logger = logging.getLogger(__name__)


class Notifications:
    """
    This object contains all functions needed for managing
    notifications and starting notification scripts.
    """

    def __init__(
        self,
        notification_file_prefix,
        max_parallel_notifications=10,
        notifications_timeout=0,
    ):
        """
        This function initialized the Notifications class.
        """
        self._active_notifications = []
        self._pending_notifications = []
        self._max_parallel_notifications = max_parallel_notifications
        self._notifications_timeout = notifications_timeout
        self._notifications_fn = os.path.join(
            notification_file_prefix, NOTIFICATION_FILE
        )
        self._notifications_log = None
        self._notifications = {}

        # Open notifications' log file
        try:
            self._notifications_log = open(self._notifications_fn, "a")
        except OSError:
            logger.critical("cannot create notifications' log file... exiting...")
            sys.exit(1)

    def has_pending_notifications(self):
        """
        This function returns True if we have pending notifications.
        """
        return len(self._pending_notifications) > 0

    def has_active_notifications(self):
        """
        This function returns True if we have active notifications.
        """
        return len(self._active_notifications) > 0

    def terminate_notification(self, my_entry):
        """
        This function terminates a notification process, and cleans up its
        output/error files.
        """
        my_p = my_entry["subp"]
        my_pid = my_entry["pid"]
        my_notification = my_entry["notification"]
        my_out_fn = my_entry["out_fn"]
        my_err_fn = my_entry["err_fn"]
        my_entry["out_fd"]
        my_entry["err_fd"]
        my_action = my_entry["action"]
        my_p.poll()
        # If process hasn't finished...
        if my_p.returncode is None:
            # Send SIGTERM first...
            try:
                os.kill(my_pid, signal.SIGTERM)
            except OSError:
                logger.info("error sending SIGTERM to notification script...")
            # Wait for child to finish
            logger.warning(
                "waiting for notification process to finish: %s - %s"
                % (my_notification, my_action)
            )
            time.sleep(WAIT_CHILD_FINISH)
            my_p.poll()
            if my_p.returncode is None:
                # Send SIGKILL now...
                logger.warning(
                    "killing notification process to finish: %s - %s"
                    % (my_notification, my_action)
                )
                try:
                    os.kill(my_pid, signal.SIGKILL)
                except OSError:
                    logger.info("error sending SIGKILL to notification script...")

        # Finally, clean up files...
        try:
            os.unlink(my_out_fn)
            os.unlink(my_err_fn)
        except OSError:
            # No error here...
            pass
        logger.warning(
            "notification terminated: {} - {}".format(my_notification, my_action)
        )

    def service_notifications(self):
        """
        This function services notifications. It chekcs the notifications
        in the active list to see if they have finished. If so, it copies
        the stdout/stderr from these notifications to the
        monitord-notifications.log file. For notifications in the
        pending_notifications list, it starts the notification scripts,
        unless there are already too many notifications running in the
        system.
        """

        # Step 1: Look at existing notifications
        if len(self._active_notifications) > 0:
            logger.info("active notifications: %d" % len(self._active_notifications))
            # We have active notifications, let's check on their statuses
            my_notif_index = 0
            while my_notif_index < len(self._active_notifications):
                my_active_notif = self._active_notifications[my_notif_index]
                # Get subprocess object
                my_active_p = my_active_notif["subp"]
                my_status = my_active_p.poll()
                if my_status is not None:
                    # Process finished notification
                    my_finished_out_fn = my_active_notif["out_fn"]
                    my_finished_err_fn = my_active_notif["err_fn"]
                    my_finished_out_fd = my_active_notif["out_fd"]
                    my_finished_err_fd = my_active_notif["err_fd"]
                    my_finished_notification = my_active_notif["notification"]
                    my_finished_action = my_active_notif["action"]
                    my_finished_notification_params = my_active_notif["params"]
                    # Close out/err files, if not already closed...
                    try:
                        my_finished_out_fd.close()
                    except OSError:
                        logger.warning(
                            "error closing stdout file for notification %s... continuing..."
                            % (my_finished_notification)
                        )
                    try:
                        my_finished_err_fd.close()
                    except OSError:
                        logger.warning(
                            "error closing stderr file for notification %s... continuing..."
                            % (my_finished_notification)
                        )

                    if self._notifications_log is not None:
                        if logger.isEnabledFor(logging.INFO):
                            self._notifications_log.write("%s\n" % ("-" * 80))
                            self._notifications_log.write(
                                "Notification time  : %s\n" % (utils.isodate())
                            )
                            self._notifications_log.write(
                                "Notification event : %s\n" % (my_finished_notification)
                            )
                            self._notifications_log.write(
                                "Notification action: %s\n" % (my_finished_action)
                            )
                            self._notifications_log.write(
                                "Notification status: %s\n" % (my_status)
                            )
                            self._notifications_log.write("\n")
                            self._notifications_log.write("Notification environment\n")
                            for k in my_finished_notification_params:
                                self._notifications_log.write(
                                    "%s : %s\n"
                                    % (k, my_finished_notification_params[k])
                                )
                            self._notifications_log.write("\n")
                            self._notifications_log.write("stdout:\n")
                            try:
                                my_f = open(my_finished_out_fn)
                                for line in my_f:
                                    self._notifications_log.write(line)
                            except OSError:
                                logger.warning(
                                    "error processing notification stdout file: %s. continuing..."
                                    % (my_finished_out_fn)
                                )
                            else:
                                my_f.close()
                            self._notifications_log.write("\n")
                            self._notifications_log.write("stderr:\n")
                            try:
                                my_f = open(my_finished_err_fn)
                                for line in my_f:
                                    self._notifications_log.write(line)
                            except OSError:
                                logger.warning(
                                    "error processing notification stderr file: %s. continuing..."
                                    % (my_finished_err_fn)
                                )
                            else:
                                my_f.close()
                            self._notifications_log.write("\n")
                            self._notifications_log.write("\n")
                        else:
                            # Only log a one-liner so we can debug things later if we need to
                            self._notifications_log.write(
                                "%s - %s - %s - %s\n"
                                % (
                                    utils.isodate(),
                                    my_finished_notification,
                                    my_finished_action,
                                    my_status,
                                )
                            )
                    else:
                        logger.critical(
                            "notifications' output log file not initialized... exiting..."
                        )
                        sys.exit(1)

                    # Now, delete output and error files
                    try:
                        os.unlink(my_finished_out_fn)
                    except OSError:
                        logger.warning(
                            "error deleting notification stdout file: %s. continuing..."
                            % (my_finished_out_fn)
                        )
                    try:
                        os.unlink(my_finished_err_fn)
                    except OSError:
                        logger.warning(
                            "error deleting notification stderr file: %s. continuing..."
                            % (my_finished_err_fn)
                        )

                    # Delete this notification from our list
                    self._active_notifications.pop(my_notif_index)
                else:
                    # Process still going... leave it...
                    my_notif_index = my_notif_index + 1

        # Step 2: Look at our notification queue
        while len(self._pending_notifications) > 0:
            # Ok we have notifications to service...
            # print "pending notifications: %s" % (len(self._pending_notifications))
            logger.info(
                "pending notifications: %s" % (len(self._pending_notifications))
            )

            # Check if we have reached the maximum number of concurrent notifications
            if len(self._active_notifications) > self._max_parallel_notifications:
                # print "reaching maximum number of concurrent notifications... waiting until next cycle..."
                logger.info(
                    "reaching maximum number of concurrent notifications... waiting until next cycle..."
                )
                break

            # Get first notification from the list
            try:
                my_action, my_env = self._pending_notifications.pop(0)
            except IndexError:
                logger.error("error processing notification list... exiting!")
                sys.exit(1)

            # Merge default environment with notification-specific environment
            my_complete_env = os.environ.copy()
            my_complete_env.update(my_env)
            try:
                my_notification = "{} - {}".format(
                    my_env["PEGASUS_JOBID"], my_env["PEGASUS_EVENT"],
                )
            except KeyError:
                logger.warning(
                    "notification missing PEGASUS_JOBID or PEGASUS_EVENT... skipping..."
                )
                continue

            # Split arguments
            my_args = shlex.split(my_action)

            # Create output and error files for the notification script to use
            try:
                my_temp_out = tempfile.mkstemp(
                    prefix="notification-", suffix="-out.log", dir="/tmp"
                )
                my_temp_err = tempfile.mkstemp(
                    prefix="notification-", suffix="-err.log", dir="/tmp"
                )
                os.close(my_temp_out[0])
                os.close(my_temp_err[0])
                my_out_fn = my_temp_out[1]
                my_err_fn = my_temp_err[1]
            except OSError:
                logger.warning(
                    "cannot create temp files for notification: %s... skipping..."
                    % (my_notification)
                )
                continue

            # Open output and error files for the notification script
            try:
                my_f_out = open(my_out_fn, "w")
                my_f_err = open(my_err_fn, "w")
            except OSError:
                logger.warning(
                    "cannot open temp files for notification: %s... skipping..."
                    % (my_notification)
                )
                try:
                    os.unlink(my_out_fn)
                    os.unlink(my_err_fn)
                except OSError:
                    # No error here...
                    pass
                continue

            # Ok, here we go...
            try:
                my_p = subprocess.Popen(
                    my_args, stdout=my_f_out, stderr=my_f_err, env=my_complete_env
                )
            except OSError:
                logger.warning(
                    "cannot start notification executable: %s... skipping..."
                    % (my_notification)
                )
                try:
                    my_f_out.close()
                    my_f_err.close()
                    os.unlink(my_out_fn)
                    os.unlink(my_err_fn)
                except OSError:
                    logger.warning(
                        "found problem cleaning up notification: %s... skipping..."
                        % (my_notification)
                    )
                    continue
                # Clean up ok, just continue
                continue
            except Exception:
                logger.warning(
                    "problem starting notification: %s... skipping..."
                    % (my_notification)
                )
                try:
                    my_f_out.close()
                    my_f_err.close()
                    os.unlink(my_out_fn)
                    os.unlink(my_err_fn)
                except OSError:
                    logger.warning(
                        "found problem cleaning up notification: %s... skipping..."
                        % (my_notification)
                    )
                    continue
                # Clean up ok, just continue
                continue

            # Let's keep everything we need for the future
            my_started_notification = {}
            my_started_notification["pid"] = my_p.pid
            my_started_notification["subp"] = my_p
            my_started_notification["env"] = my_complete_env
            my_started_notification["params"] = my_env
            my_started_notification["args"] = my_args
            my_started_notification["action"] = my_action
            my_started_notification["out_fd"] = my_f_out
            my_started_notification["err_fd"] = my_f_err
            my_started_notification["out_fn"] = my_out_fn
            my_started_notification["err_fn"] = my_err_fn
            my_started_notification["notification"] = my_notification
            my_started_notification["time"] = time.time()

            # Add to the active list, and done!
            self._active_notifications.append(my_started_notification)
            logger.info("started notification for: %s" % (my_notification))

        # Step 3: Check if any notifications ran over the allowed time
        if self._notifications_timeout > 0:
            # Only go through the list if a timeout was specified

            # Get current time
            now = int(math.floor(time.time()))

            # Go through our list
            my_index = 0
            while my_index < len(self._active_notifications):
                my_entry = self._active_notifications[my_index]
                my_exp_time = my_entry["time"] + self._notifications_timeout

                # Check if notification has expired
                if my_exp_time < now:
                    # Notification has expired... kill it...
                    logger.warning("notification expired... terminating it...")
                    self.terminate_notification(my_entry)
                    # Delete this notification from our list
                    self._active_notifications.pop(my_index)
                else:
                    # Notification hasn't expired yet, move to next one...
                    my_index = my_index + 1

    def finish_notifications(self):
        """
        This function flushes all notifications, and closes the
        notifications' log file. It also logs all pending (but not yet
        issued) notifications.
        """
        # Take care of active notifications
        if len(self._active_notifications) > 0:
            for my_entry in self._active_notifications:
                self.terminate_notification(my_entry)

        # Take care of pending notifications
        if len(self._pending_notifications) > 0:
            for my_action, my_env in self._pending_notifications:
                try:
                    my_notification = "{} - {}".format(
                        my_env["PEGASUS_JOBID"], my_env["PEGASUS_EVENT"],
                    )
                except KeyError:
                    logger.warning(
                        "notification missing PEGASUS_JOBID or PEGASUS_EVENT... skipping..."
                    )
                    continue
                logger.warning(
                    "pending notification skipped: %s - %s"
                    % (my_notification, my_action)
                )

        # Close notifications' log file
        if self._notifications_log is not None:
            try:
                self._notifications_log.close()
            except OSError:
                logger.warning("error closing notifications' log file...")
            self._notifications_log = None

    def read_notification_file(self, notify_file, wf_uuid):
        """
        This function reads the notification file, parsing all
        notifications and creating our list of events to track.
        It returns the number of notifications read from the
        notifications' file.
        """
        if notify_file is None:
            return 0

        logger.info("loading notifications from %s" % (notify_file))

        # Open file
        try:
            NOTIFY = open(notify_file)
        except OSError:
            logger.warning(
                "cannot load notification file %s, continuing without notifications"
                % (notify_file)
            )
            return 0

        # Start with empty dictionaries for the three types of notifications
        my_notifications_read = 0
        my_notifications = {"workflow": {}, "job": {}, "invocation": {}}
        # For workflow and job notifications, we have a dict(workflow_id|job_id, dict(cond, [actions]))
        # For invocation notifications, we have a dict(job_id, dict(inv_id, dict(cond, [actions])))

        # Process notifications
        for line in NOTIFY:
            line = line.strip()
            # Skip blank lines
            if len(line) == 0:
                continue
            # Skip comments
            if line.startswith("#"):
                continue

            # Check if we split it in 4 or 5 pieces
            if line.lower().startswith("invocation"):
                # This is an invocation notification, split and get all pieces
                my_entry = line.split(None, 4)
                if len(my_entry) != 5:
                    logger.warning(
                        "cannot parse notification: %s, skipping..." % (line)
                    )
                    continue
                my_type = my_entry[0].lower()
                my_id = my_entry[1]
                try:
                    my_inv = int(my_entry[2])
                except ValueError:
                    logger.warning(
                        "cannot parse notification: %s, skipping..." % (line)
                    )
                    continue
                my_condition = my_entry[3]
                my_action = my_entry[4]
            else:
                # This is a workflow/job notification, split and get all pieces
                my_entry = line.split(None, 3)
                if len(my_entry) != 4:
                    logger.warning(
                        "cannot parse notification: %s, skipping..." % (line)
                    )
                    continue
                my_type = my_entry[0].lower()
                my_id = my_entry[1]
                my_condition = my_entry[2]
                my_action = my_entry[3]

            # Pick the right dictionary, depending on event type
            if my_type == "workflow":
                my_dict = my_notifications["workflow"]
                if my_id != wf_uuid:
                    logger.warning(
                        "workflow notification has id %s, our id is %s, skipping..."
                        % (my_id, wf_uuid)
                    )
                    continue

            elif my_type == "job" or my_type == "daxjob" or my_type == "dagjob":
                my_dict = my_notifications["job"]
            elif my_type == "invocation":
                my_dict = my_notifications["invocation"]
            else:
                logger.warning("unknown notification type: %s, skipping..." % (line))
                continue

            logger.debug("loading notification: %s" % (line))
            my_notifications_read = my_notifications_read + 1

            # Make sure id is in dictionary
            if not my_id in my_dict:
                my_dict[my_id] = {}

            # For invocations, one extra level...
            if my_type == "invocation":
                my_dict = my_dict[my_id]
                if not my_inv in my_dict:
                    my_dict[my_inv] = {}
                # Now add the notification condition, action pair
                if not my_condition in my_dict[my_inv]:
                    # No actions, start with the list
                    my_dict[my_inv][my_condition] = [my_action]
                else:
                    # We already have an action(s), let's add the new one to the list
                    my_dict[my_inv][my_condition].append(my_action)
            else:
                # Now add the notification condition, action pair
                if not my_condition in my_dict[my_id]:
                    my_dict[my_id][my_condition] = [my_action]
                else:
                    my_dict[my_id][my_condition].append(my_action)

        # Save our notifications for later use...
        if wf_uuid in self._notifications:
            logger.debug("reloaded notifications for workflow %s" % (wf_uuid))
        self._notifications[wf_uuid] = my_notifications

        # Close file
        try:
            NOTIFY.close()
        except OSError:
            pass

        # Return number of notifications read
        logger.debug(
            "loaded %d notifications for workflow %s" % (my_notifications_read, wf_uuid)
        )
        return my_notifications_read

    def process_workflow_notifications(self, wf, state):
        """
        This function takes care of processing workflow-level notifications.
        """
        # Check if we have notifications for this workflow
        if not wf._wf_uuid in self._notifications:
            return

        # Get the notifications' dictionary for this workflow id
        wf_notifications = self._notifications[wf._wf_uuid]

        if "workflow" in wf_notifications:
            my_dict = wf_notifications["workflow"]
            if len(my_dict) == 0:
                # No workflow notifications
                return
        else:
            logger.warning("notification structure missing workflow entry...")
            return

        # Our workflow is must be in there...
        if wf._wf_uuid in my_dict:
            my_notifications = my_dict[wf._wf_uuid]
        else:
            logger.warning(
                "notification has mismatching workflow id: %s different from %s"
                % (wf._wf_uuid, str(my_dict))
            )
            return

        # Sanity check the state...
        if state != "start" and state != "end":
            logger.warning("unknown workflow state %s, continuing..." % (state))
            return

        # Now, match the workflow state to the conditions in the notifications...
        for k in my_notifications:
            # Look up the actions for this notification now
            my_actions = my_notifications[k]
            if state == "start":
                if k != "start" and k != "all":
                    continue
                # Change k == 'all' to 'start'
                k = "start"
            if state == "end":
                if k == "on_error":
                    if wf._dagman_exit_code == 0:
                        continue
                elif k == "on_success":
                    if wf._dagman_exit_code != 0:
                        continue
                elif k != "at_end" and k != "all":
                    continue
                if k == "all":
                    k = "at_end"

            # Ok, we have a match!
            for action in my_actions:
                # Create dictionary with needed environment variables
                my_env = {}
                my_env["PEGASUS_EVENT"] = k
                my_env["PEGASUS_EVENT_TIMESTAMP"] = str(wf._current_timestamp)
                my_env["PEGASUS_EVENT_TIMESTAMP_ISO"] = utils.isodate(
                    wf._current_timestamp
                )
                my_env["PEGASUS_SUBMIT_DIR"] = wf._original_submit_dir
                my_env["PEGASUS_STDOUT"] = wf._out_file
                my_env["PEGASUS_JOBID"] = wf._wf_uuid
                my_env["PEGASUS_WFID"] = (
                    (wf._dax_label or "unknown") + "-" + (wf._dax_index or "unknown")
                )
                if state == "end":
                    # Workflow status is already in plain format, no need for conversion
                    my_env["PEGASUS_STATUS"] = str(wf._dagman_exit_code)

                # Done, queue the notification
                self._pending_notifications.append((action, my_env))
                # print "WORKFLOW NOTIFICATION ---> ", action, my_env

    def process_job_notifications(self, wf, state, job, status):
        """
        This function takes care of processing job-level notifications.
        """
        # Check if we have notifications for this workflow
        if not wf._wf_uuid in self._notifications:
            return

        # Get the notifications' dictionary for this workflow id
        wf_notifications = self._notifications[wf._wf_uuid]

        if "job" in wf_notifications:
            my_dict = wf_notifications["job"]
        else:
            logger.warning("notification structure missing job entry...")
            return

        # Check if we have notifications for this job
        if not job._exec_job_id in my_dict:
            return

        my_notifications = my_dict[job._exec_job_id]
        if job._exec_job_id in wf._job_info:
            if wf._job_info[job._exec_job_id][3] is None:
                job_has_post_script = False
            else:
                job_has_post_script = True
        else:
            logger.warning(
                "cannot find job %s in job_info database... skipping notification..."
                % (job._exec_job_id)
            )
            return

        # Now, match the job state to the conditions in the notifications...
        for k in my_notifications:
            # Look up the actions for this notification now
            my_actions = my_notifications[k]
            if state == "EXECUTE":
                if k != "start" and k != "all":
                    continue
                # Change k to "start"
                k = "start"
                my_status = None
            elif state == "JOB_SUCCESS":
                if job_has_post_script:
                    # Wait till postscript...
                    continue
                if k == "start" or k == "on_error":
                    continue
                if k == "all":
                    k = "at_end"
                my_status = "0"
            elif state == "POST_SCRIPT_SUCCESS":
                if k == "start" or k == "on_error":
                    continue
                if k == "all":
                    k = "at_end"
                my_status = "0"
            elif state == "JOB_FAILURE":
                if job_has_post_script:
                    # Wait till postscript...
                    continue
                if k == "start" or k == "on_success":
                    continue
                if k == "all":
                    k = "at_end"
                my_status = status
            elif state == "POST_SCRIPT_FAILURE":
                if k == "start" or k == "on_success":
                    continue
                if k == "all":
                    k = "at_end"
                my_status = status
            else:
                # We are in some other state...
                continue

            my_output = os.path.join(wf._original_submit_dir, job._output_file)
            my_error = os.path.join(wf._original_submit_dir, job._error_file)
            # Use the rotated file names if at the end of the job
            if k != "start":
                my_output = my_output + ".%03d" % (job._job_output_counter)
                my_error = my_error + ".%03d" % (job._job_output_counter)

            # Ok, we have a match!
            for action in my_actions:
                # Create dictionary with needed environment variables
                my_env = {}
                my_env["PEGASUS_EVENT"] = k
                my_env["PEGASUS_EVENT_TIMESTAMP"] = str(wf._current_timestamp)
                my_env["PEGASUS_EVENT_TIMESTAMP_ISO"] = utils.isodate(
                    wf._current_timestamp
                )
                my_env["PEGASUS_SUBMIT_DIR"] = wf._original_submit_dir
                my_env["PEGASUS_JOBID"] = job._exec_job_id
                my_env["PEGASUS_WFID"] = (
                    (wf._dax_label or "unknown") + "-" + (wf._dax_index or "unknown")
                )
                my_env["PEGASUS_STDOUT"] = my_output
                my_env["PEGASUS_STDERR"] = my_error
                if my_status is not None:
                    my_env["PEGASUS_STATUS"] = str(my_status)

                # Done, queue the notification
                self._pending_notifications.append((action, my_env))
                # print "JOB NOTIFICATION ---> ", action, my_env

    def process_invocation_notifications(self, wf, job, task_id, record=None):
        """
        This function takes care of processing invocation-level notifications.
        """
        if record is None:
            record = {}

        # Check if we have notifications for this workflow
        if not wf._wf_uuid in self._notifications:
            return

        # Get the notifications' dictionary for this workflow id
        wf_notifications = self._notifications[wf._wf_uuid]

        if "invocation" in wf_notifications:
            my_dict = wf_notifications["invocation"]
        else:
            logger.warning("notification structure missing invocation entry...")
            return

        # Check if we have notifications for this job
        if not job._exec_job_id in my_dict:
            return

        # Advance to the task dictionary
        my_dict = my_dict[job._exec_job_id]

        # Check if we have notifications for this invocation
        if not task_id in my_dict:
            return

        my_notifications = my_dict[task_id]

        # Now, match the invocation state to the condition in the notification
        for k in my_notifications:
            # Look up the actions for this notification now
            my_actions = my_notifications[k]
            if "raw" in record:
                my_status = record["raw"]
            else:
                my_status = job._main_job_exitcode
            # Convert exitcode to int
            try:
                my_status = int(my_status)
            except ValueError:
                pass
            # Now, compare to the notification condition(s)
            if my_status == 0:
                if k == "on_error":
                    continue
            if my_status != 0:
                if k == "on_success":
                    continue
            if k == "all":
                k = "at_end"

            # Here, we always use the rotated file names as the invocation has already finished...
            my_output = os.path.join(
                wf._original_submit_dir, job._output_file
            ) + ".%03d" % (job._job_output_counter)
            my_error = os.path.join(
                wf._original_submit_dir, job._error_file
            ) + ".%03d" % (job._job_output_counter)

            # Ok, we have a match!
            for action in my_actions:
                # Create dictionary with needed environment variables
                my_env = {}
                my_env["PEGASUS_EVENT"] = k
                my_env["PEGASUS_EVENT_TIMESTAMP"] = str(wf._current_timestamp)
                my_env["PEGASUS_EVENT_TIMESTAMP_ISO"] = utils.isodate(
                    wf._current_timestamp
                )
                my_env["PEGASUS_SUBMIT_DIR"] = wf._original_submit_dir
                my_env["PEGASUS_JOBID"] = job._exec_job_id
                my_env["PEGASUS_INVID"] = str(task_id)
                my_env["PEGASUS_WFID"] = (
                    (wf._dax_label or "unknown") + "-" + (wf._dax_index or "unknown")
                )
                my_env["PEGASUS_STDOUT"] = my_output
                my_env["PEGASUS_STDERR"] = my_error
                if k != "start":
                    # Convert raw exitcode into human-parseable format
                    my_env["PEGASUS_STATUS"] = str(utils.raw_to_regular(my_status))

                # Done, queue the notification
                self._pending_notifications.append((action, my_env))
                # print "INVOCATION NOTIFICATION ---> ", action, my_env

    def remove_notifications(self, wf_uuid):
        """
        This function removes the notifications for workflow wf_uuid
        from our _notifications dictionary.
        """

        # Check if we have notifications for this workflow
        if not wf_uuid in self._notifications:
            return

        logger.debug("deleting notifications for workflow %s..." % (wf_uuid))

        # Delete them from our dictionary
        del self._notifications[wf_uuid]



__author__ = "Rajiv Mayani"

import logging
import uuid
from decimal import Decimal
from enum import Enum

from flask.json import JSONEncoder
from sqlalchemy import inspect

from Pegasus.service.base import ErrorResponse, OrderedDict, PagedResponse

log = logging.getLogger(__name__)


class PegasusJsonEncoder(JSONEncoder):
    """JSON Encoder for Pegasus Service API Resources."""

    def default(self, o):
        """."""

        if isinstance(o, uuid.UUID):
            return str(o)

        elif isinstance(o, Decimal):
            return float(o)

        elif isinstance(o, Enum):
            return o.name

        elif isinstance(o, PagedResponse):
            json_record = OrderedDict([("records", o.records)])

            if o.total_records or o.total_filtered:
                meta = OrderedDict()

                if o.total_records is not None:
                    meta["records_total"] = o.total_records

                if o.total_filtered is not None:
                    meta["records_filtered"] = o.total_filtered

                json_record["_meta"] = meta

            return json_record

        elif isinstance(o, ErrorResponse):
            json_record = OrderedDict([("code", o.code), ("message", o.message)])

            if o.errors:
                json_record["errors"] = [{"field": f, "errors": e} for f, e in o.errors]

            return json_record

        elif hasattr(o, "__json__"):
            return o.__json__()

        elif hasattr(o, "__table__"):
            unloaded = inspect(o).unloaded
            _v = {
                k: getattr(o, k)
                for k in o.__mapper__.column_attrs.keys()
                if k not in unloaded
            }

            for k in getattr(o, "__includes__", {}):
                _v[k] = getattr(o, k)

            for k in getattr(o, "__excludes__", {}):
                del _v[k]

            return _v

        return JSONEncoder.default(self, o)



#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import logging
import warnings

from sqlalchemy import *
from sqlalchemy.exc import *

from Pegasus.db.admin.versions.base_version import BaseVersion
from Pegasus.db.schema import *

DB_VERSION = 5

log = logging.getLogger(__name__)


class Version(BaseVersion):
    def __init__(self, connection):
        super().__init__(connection)

    def update(self, force=False):
        "Fixes malfunction past migrations and clean the database"
        log.info("Updating to version %s" % DB_VERSION)

        # check and fix foreign key for master_workflowstate table
        with warnings.catch_warnings():
            table_names = self.db.get_bind().table_names()
            tables_to_inquiry = ["master_workflowstate", "master_workflow", "workflow"]
            if set(tables_to_inquiry).issubset(table_names):
                meta = MetaData()
                warnings.simplefilter("ignore")
                meta.reflect(bind=self.db.get_bind(), only=tables_to_inquiry)
                mw_id = meta.tables["master_workflowstate"].c.wf_id

                # PM-1015: invalid constraint
                if not mw_id.references(
                    meta.tables["master_workflow"].c.wf_id
                ) and mw_id.references(meta.tables["workflow"].c.wf_id):
                    log.info("Updating foreign key constraint.")
                    try:
                        self.db.execute("DROP INDEX UNIQUE_MASTER_WORKFLOWSTATE")
                    except Exception:
                        pass
                    if self.db.get_bind().driver == "mysqldb":
                        self.db.execute(
                            "RENAME TABLE master_workflowstate TO master_workflowstate_v4"
                        )
                    else:
                        self.db.execute(
                            "ALTER TABLE master_workflowstate RENAME TO master_workflowstate_v4"
                        )
                    Workflowstate.__table__.create(self.db.get_bind(), checkfirst=True)
                    self.db.execute(
                        "INSERT INTO master_workflowstate(wf_id, state, timestamp, restart_count, status) SELECT m4.wf_id, m4.state, m4.timestamp, m4.restart_count, m4.status FROM master_workflowstate_v4 m4 LEFT JOIN master_workflow mw WHERE m4.wf_id=mw.wf_id"
                    )
                    self.db.commit()
                    self._drop_table("master_workflowstate_v4")

        # clean the database from past migrations
        self._drop_table("rc_lfn_new")
        self._drop_table("rc_lfn_old")

    def downgrade(self, force=False):
        "No need for downgrade"

    def _drop_table(self, table_name):
        """
        Drop a table.
        :param table_name:
        :return:
        """
        try:
            self.db.execute("DROP TABLE %s" % table_name)
        except Exception:
            pass



import logging
import os

from flask import g, redirect, render_template, request, send_from_directory, url_for
from sqlalchemy.orm.exc import NoResultFound

from Pegasus.db.admin.admin_loader import DBAdminError
from Pegasus.db.errors import StampedeDBNotFoundError
from Pegasus.service import filters
from Pegasus.service._serialize import serialize
from Pegasus.service.base import ErrorResponse, ServiceError
from Pegasus.service.dashboard import blueprint
from Pegasus.service.dashboard.dashboard import Dashboard, NoWorkflowsFoundError
from Pegasus.service.dashboard.queries import MasterDBNotFoundError
from Pegasus.tools import utils

log = logging.getLogger(__name__)


@blueprint.route("/")
def redirect_to_index():
    return redirect(url_for(".index"))


@blueprint.route("/u/<username>/")
def index(username):
    """
    List all workflows from the master database.
    """
    try:
        dashboard = Dashboard(g.master_db_url)
        args = __get_datatables_args()
        if request.is_xhr:
            count, filtered, workflows, totals = dashboard.get_root_workflow_list(
                **args
            )
        else:
            totals = dashboard.get_root_workflow_list(counts_only=True, **args)

    except NoWorkflowsFoundError as e:
        if request.is_xhr:
            workflows = []
            d = {
                "draw": args["sequence"] if args["sequence"] else 0,
                "recordsTotal": e.count if e.count is not None else len(workflows),
                "data": workflows,
            }
            if args["limit"]:
                d["recordsFiltered"] = e.filtered

            return serialize(d)

        return render_template("workflow.html", counts=(0, 0, 0, 0))

    if request.is_xhr:
        d = {
            "draw": args["sequence"] if args["sequence"] else 0,
            "recordsTotal": count if count is not None else len(workflows),
            "data": [w._asdict() for w in workflows],
        }
        if args["limit"]:
            d["recordsFiltered"] = filtered

        return serialize(d)

    return render_template("workflow.html", counts=totals)


@blueprint.route("/u/<username>/r/<root_wf_id>/w")
@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>")
def workflow(username, root_wf_id, wf_id=None):
    """
    Get details for a specific workflow.
    """
    wf_uuid = request.args.get("wf_uuid", None)

    if not wf_id and not wf_uuid:
        raise ValueError("Workflow ID or Workflow UUID is required")

    if wf_id:
        dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id=wf_id)
    else:
        dashboard = Dashboard(g.master_db_url, root_wf_id)

    try:
        counts, details, statistics = dashboard.get_workflow_information(wf_id, wf_uuid)
    except NoResultFound:
        return render_template("error/workflow/workflow_details_missing.html")

    return render_template(
        "workflow/workflow_details.html",
        root_wf_id=root_wf_id,
        wf_id=details.wf_id,
        workflow=details,
        counts=counts,
        statistics=statistics,
    )


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/sw/", methods=["GET"])
def sub_workflows(username, root_wf_id, wf_id):
    """
    Get a list of all sub-workflow of a given workflow.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    sub_workflows = dashboard.get_sub_workflows(wf_id)

    for i in range(len(sub_workflows)):
        sub_workflows[i] = sub_workflows[i]._asdict()

    return serialize(sub_workflows)


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/j/failed/", methods=["GET"])
def failed_jobs(username, root_wf_id, wf_id):
    """
    Get a list of all failed jobs of the latest instance for a given workflow.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    args = __get_datatables_args()

    total_count, filtered_count, failed_jobs_list = dashboard.get_failed_jobs(
        wf_id, **args
    )

    for i in range(len(failed_jobs_list)):
        failed_jobs_list[i] = failed_jobs_list[i]._asdict()
        failed_jobs_list[i]["DT_RowClass"] = "failing"

    d = {
        "draw": args["sequence"] if args["sequence"] else 0,
        "recordsTotal": total_count
        if total_count is not None
        else len(failed_jobs_list),
        "data": failed_jobs_list,
    }
    if args["limit"]:
        d["recordsFiltered"] = filtered_count

    return serialize(d)


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/j/running/", methods=["GET"])
def running_jobs(username, root_wf_id, wf_id):
    """
    Get a list of all running jobs of the latest instance for a given workflow.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    args = __get_datatables_args()

    total_count, filtered_count, running_jobs_list = dashboard.get_running_jobs(
        wf_id, **args
    )

    for i in range(len(running_jobs_list)):
        running_jobs_list[i] = running_jobs_list[i]._asdict()
        running_jobs_list[i]["DT_RowClass"] = "running"

    d = {
        "draw": args["sequence"] if args["sequence"] else 0,
        "recordsTotal": total_count
        if total_count is not None
        else len(running_jobs_list),
        "data": running_jobs_list,
    }
    if args["limit"]:
        d["recordsFiltered"] = filtered_count

    return serialize(d)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/successful/", methods=["GET"]
)
def successful_jobs(username, root_wf_id, wf_id):
    """
    Get a list of all successful jobs of the latest instance for a given workflow.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    args = __get_datatables_args()

    total_count, filtered_count, successful_jobs_list = dashboard.get_successful_jobs(
        wf_id, **args
    )

    for i in range(len(successful_jobs_list)):
        successful_jobs_list[i] = successful_jobs_list[i]._asdict()
        successful_jobs_list[i]["DT_RowClass"] = "successful"
        successful_jobs_list[i]["root_wf_id"] = root_wf_id
        successful_jobs_list[i]["wf_id"] = wf_id
        successful_jobs_list[i]["duration_formatted"] = filters.time_to_str(
            successful_jobs_list[i]["duration"]
        )

    d = {
        "draw": args["sequence"] if args["sequence"] else 0,
        "recordsTotal": total_count
        if total_count is not None
        else len(successful_jobs_list),
        "data": successful_jobs_list,
    }
    if args["limit"]:
        d["recordsFiltered"] = filtered_count

    return serialize(d)


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/j/failing/", methods=["GET"])
def failing_jobs(username, root_wf_id, wf_id):
    """
    Get a list of failing jobs of the latest instance for a given workflow.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    args = __get_datatables_args()

    total_count, filtered_count, failing_jobs_list = dashboard.get_failing_jobs(
        wf_id, **args
    )

    for i in range(len(failing_jobs_list)):
        failing_jobs_list[i] = failing_jobs_list[i]._asdict()
        failing_jobs_list[i]["DT_RowClass"] = "failing"

    d = {
        "draw": args["sequence"] if args["sequence"] else 0,
        "recordsTotal": total_count
        if total_count is not None
        else len(failing_jobs_list),
        "data": failing_jobs_list,
    }
    if args["limit"]:
        d["recordsFiltered"] = filtered_count

    return serialize(d)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/<job_id>/ji/<job_instance_id>",
    methods=["GET"],
)
def job(username, root_wf_id, wf_id, job_id, job_instance_id):
    """
    Get details of a specific job instance.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    job = dashboard.get_job_information(wf_id, job_id, job_instance_id)
    job_states = dashboard.get_job_states(wf_id, job_id, job_instance_id)
    job_instances = dashboard.get_job_instances(wf_id, job_id)

    if not job:
        return "Bad Request", 400

    return render_template(
        "workflow/job/job_details.html",
        root_wf_id=root_wf_id,
        wf_id=wf_id,
        job_id=job_id,
        job=job,
        job_instances=job_instances,
        job_states=job_states,
    )


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/<job_id>/ji/<job_instance_id>/stdout",
    methods=["GET"],
)
def stdout(username, root_wf_id, wf_id, job_id, job_instance_id):
    """
    Get stdout contents for a specific job instance.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    text = dashboard.get_stdout(wf_id, job_id, job_instance_id)

    if text.stdout_text is None:
        return "No stdout for workflow " + wf_id + " job-id " + job_id
    else:
        return "<pre>%s</pre>" % utils.unquote(text.stdout_text)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/<job_id>/ji/<job_instance_id>/stderr",
    methods=["GET"],
)
def stderr(username, root_wf_id, wf_id, job_id, job_instance_id):
    """
    Get stderr contents for a specific job instance.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    text = dashboard.get_stderr(wf_id, job_id, job_instance_id)

    if text.stderr_text is None:
        return "No Standard error for workflow " + wf_id + " job-id " + job_id
    else:
        return "<pre>%s</pre>" % utils.unquote(text.stderr_text)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/<job_id>/ji/<job_instance_id>/i/successful",
    methods=["GET"],
)
def successful_invocations(username, root_wf_id, wf_id, job_id, job_instance_id):
    """
    Get list of successful invocations for a given job.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    successful_invocations_list = dashboard.get_successful_job_invocation(
        wf_id, job_id, job_instance_id
    )

    for i in range(len(successful_invocations_list)):
        successful_invocations_list[i] = successful_invocations_list[i]._asdict()
        successful_invocations_list[i]["remote_duration"] = filters.time_to_str(
            successful_invocations_list[i]["remote_duration"]
        )

    return serialize(successful_invocations_list)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/<job_id>/ji/<job_instance_id>/i/failed",
    methods=["GET"],
)
def failed_invocations(username, root_wf_id, wf_id, job_id, job_instance_id):
    """
    Get list of failed invocations for a given job.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    failed_invocations_list = dashboard.get_failed_job_invocation(
        wf_id, job_id, job_instance_id
    )

    for i in range(len(failed_invocations_list)):
        failed_invocations_list[i] = failed_invocations_list[i]._asdict()
        failed_invocations_list[i]["remote_duration"] = filters.time_to_str(
            failed_invocations_list[i]["remote_duration"]
        )

    return serialize(failed_invocations_list)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/<job_id>/ji/<job_instance_id>/i/",
    methods=["GET"],
)
@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/j/<job_id>/ji/<job_instance_id>/i/<invocation_id>",
    methods=["GET"],
)
def invocation(username, root_wf_id, wf_id, job_id, job_instance_id, invocation_id):
    """
    Get detailed invocation information
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    invocation = dashboard.get_invocation_information(
        wf_id, job_id, job_instance_id, invocation_id
    )

    return render_template(
        "workflow/job/invocation/invocation_details.html",
        root_wf_id=root_wf_id,
        wf_id=wf_id,
        job_id=job_id,
        job_instance_id=job_instance_id,
        invocation_id=invocation_id,
        invocation=invocation,
    )


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/charts", methods=["GET"])
def charts(username, root_wf_id, wf_id):
    """
    Get job-distribution information
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    job_dist = dashboard.plots_transformation_statistics(wf_id)

    d = []
    for i in range(len(job_dist)):
        d.append(
            {
                "name": job_dist[i].transformation,
                "count": {
                    "total": job_dist[i].count,
                    "success": job_dist[i].success,
                    "failure": job_dist[i].failure,
                },
                "time": {
                    "total": job_dist[i].sum,
                    "min": job_dist[i].min,
                    "max": job_dist[i].max,
                    "avg": job_dist[i].avg,
                },
            }
        )

    return render_template(
        "workflow/charts.html", root_wf_id=root_wf_id, wf_id=wf_id, job_dist=d
    )


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/charts/time_chart", methods=["GET"]
)
def time_chart(username, root_wf_id, wf_id):
    """
    Get job-distribution information
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    time_chart_job, time_chart_invocation = dashboard.plots_time_chart(wf_id)

    d = []
    for i in range(len(time_chart_job)):
        d.append(
            {
                "date_format": time_chart_job[i].date_format,
                "count": {
                    "job": time_chart_job[i].count,
                    "invocation": time_chart_invocation[i].count,
                },
                "total_runtime": {
                    "job": time_chart_job[i].total_runtime,
                    "invocation": time_chart_invocation[i].total_runtime,
                },
            }
        )

    return serialize(d)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/charts/gantt_chart", methods=["GET"]
)
def gantt_chart(username, root_wf_id, wf_id):
    """
    Get information required to generate a Gantt chart.
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    gantt_chart = dashboard.plots_gantt_chart()

    d = []
    for i in range(len(gantt_chart)):
        d.append(
            {
                "job_id": gantt_chart[i].job_id,
                "job_instance_id": gantt_chart[i].job_instance_id,
                "job_submit_seq": gantt_chart[i].job_submit_seq,
                "job_name": gantt_chart[i].job_name,
                "transformation": gantt_chart[i].transformation,
                "jobS": gantt_chart[i].jobS,
                "jobDuration": gantt_chart[i].jobDuration,
                "pre_start": gantt_chart[i].pre_start,
                "pre_duration": gantt_chart[i].pre_duration,
                "condor_start": gantt_chart[i].condor_start,
                "condor_duration": gantt_chart[i].condor_duration,
                "grid_start": gantt_chart[i].grid_start,
                "grid_duration": gantt_chart[i].grid_duration,
                "exec_start": gantt_chart[i].exec_start,
                "exec_duration": gantt_chart[i].exec_duration,
                "kickstart_start": gantt_chart[i].kickstart_start,
                "kickstart_duration": gantt_chart[i].kickstart_duration,
                "post_start": gantt_chart[i].post_start,
                "post_duration": gantt_chart[i].post_duration,
            }
        )

    return serialize(d)


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/statistics", methods=["GET"])
def statistics(username, root_wf_id, wf_id):
    """
    Get workflow statistics information
    """
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    summary_times = dashboard.workflow_summary_stats(wf_id)

    for key, value in summary_times.items():
        summary_times[key] = filters.time_to_str(value)

    workflow_stats = dashboard.workflow_stats()

    return render_template(
        "workflow/statistics.html",
        root_wf_id=root_wf_id,
        wf_id=wf_id,
        summary_stats=summary_times,
        workflow_stats=workflow_stats,
    )


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/statistics/summary", methods=["GET"]
)
def workflow_summary_stats(username, root_wf_id, wf_id):
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    summary_times = dashboard.workflow_summary_stats(wf_id)

    for key, value in summary_times.items():
        summary_times[key] = filters.time_to_str(value)

    return serialize(summary_times)


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/statistics/workflow", methods=["GET"]
)
def workflow_stats(username, root_wf_id, wf_id):
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    return serialize(dashboard.workflow_stats())


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/statistics/job_breakdown", methods=["GET"]
)
def job_breakdown_stats(username, root_wf_id, wf_id):
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    return serialize(dashboard.job_breakdown_stats())


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/statistics/job", methods=["GET"]
)
def job_stats(username, root_wf_id, wf_id):
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    return serialize(dashboard.job_stats())


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/statistics/integrity", methods=["GET"]
)
def integrity_stats(username, root_wf_id, wf_id):
    dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id)
    return serialize(dashboard.integrity_stats())


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/browser", methods=["GET"])
def file_browser(username, root_wf_id, wf_id):
    try:
        dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id=wf_id)
        details = dashboard.get_workflow_details(wf_id)
        submit_dir = details.submit_dir

        if os.path.isdir(submit_dir):
            init_file = request.args.get("init_file", None)
            return render_template(
                "file-browser.html",
                root_wf_id=root_wf_id,
                wf_id=wf_id,
                init_file=init_file,
            )
        else:
            raise ServiceError(
                ErrorResponse(
                    "SUBMIT_DIR_NOT_FOUND",
                    "%r is not a valid directory" % str(submit_dir),
                )
            )

    except NoResultFound:
        return render_template("error/workflow/workflow_details_missing.html")

    return "Error", 500


@blueprint.route("/u/<username>/r/<root_wf_id>/w/<wf_id>/files/", methods=["GET"])
@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/files/<path:path>", methods=["GET"]
)
def file_list(username, root_wf_id, wf_id, path=""):
    try:
        dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id=wf_id)
        details = dashboard.get_workflow_details(wf_id)
        submit_dir = details.submit_dir

        if os.path.isdir(submit_dir):
            dest = os.path.join(submit_dir, path)

            if os.path.isfile(dest):
                return "", 204

            folders = {"dirs": [], "files": []}

            for entry in os.listdir(dest):
                if os.path.isdir(os.path.join(dest, entry)):
                    folders["dirs"].append(os.path.normpath(os.path.join(path, entry)))
                else:
                    folders["files"].append(os.path.normpath(os.path.join(path, entry)))

            return serialize(folders), 200, {"Content-Type": "application/json"}

        else:
            raise ServiceError(
                ErrorResponse(
                    "SUBMIT_DIR_NOT_FOUND",
                    "%r is not a valid directory" % str(submit_dir),
                )
            )

    except NoResultFound:
        return render_template("error/workflow/workflow_details_missing.html")

    return "Error", 500


@blueprint.route(
    "/u/<username>/r/<root_wf_id>/w/<wf_id>/file/<path:path>", methods=["GET"]
)
def file_view(username, root_wf_id, wf_id, path):
    try:
        dashboard = Dashboard(g.master_db_url, root_wf_id, wf_id=wf_id)
        details = dashboard.get_workflow_details(wf_id)
        submit_dir = details.submit_dir

        file_path = os.path.join(submit_dir, path)
        if not os.path.isfile(file_path):
            return "File not found", 404

        return send_from_directory(submit_dir, path)
    except NoResultFound:
        return render_template("error/workflow/workflow_details_missing.html")

    return "Error", 500


@blueprint.route("/u/<username>/info")
def info(username):
    return render_template("info.html")


def __get_datatables_args():
    """
    Extract list of arguments passed in the request
    """

    table_args = dict()

    #
    # Common Arguments
    #

    table_args["column-count"] = 0
    table_args["sort-col-count"] = 0

    if request.args.get("draw"):
        table_args["sequence"] = request.args.get("draw")

    if request.args.get("start"):
        table_args["offset"] = int(request.args.get("start"))

    if request.args.get("length"):
        table_args["limit"] = int(request.args.get("length"))

    if request.args.get("search[value]"):
        table_args["filter"] = request.args.get("search[value]")

    if request.args.get("search[regex]"):
        table_args["filter-regex"] = request.args.get("search[regex]")

    #
    # Custom Arguments
    #

    if request.args.get("time_filter"):
        table_args["time_filter"] = request.args.get("time_filter")

    i = 0
    while True:
        if request.args.get("columns[%d][data]" % i):
            table_args["column-count"] += 1
            table_args["mDataProp_%d" % i] = request.args.get("columns[%d][data]" % i)
        else:
            break

        #
        # Column Search
        #

        if request.args.get("columns[%d][searchable]" % i):
            table_args["bSearchable_%d" % i] = request.args.get(
                "columns[%d][searchable]" % i
            )

        if request.args.get("columns[%d][search][value]" % i):
            table_args["sSearch_%d" % i] = request.args.get(
                "columns[%d][search][value]" % i
            )

        if request.args.get("columns[%d][search][regex]" % i):
            table_args["bRegex_%d" % i] = request.args.get(
                "columns[%d][search][regex]" % i
            )

        #
        # Column Sort
        #

        if request.args.get("columns[%d][orderable]" % i):
            table_args["bSortable_%d" % i] = request.args.get(
                "columns[%d][orderable]" % i
            )

        if request.args.get("order[%d][column]" % i):
            table_args["sort-col-count"] += 1
            table_args["iSortCol_%d" % i] = int(
                request.args.get("order[%d][column]" % i)
            )

        if request.args.get("order[%d][dir]" % i):
            table_args["sSortDir_%d" % i] = request.args.get("order[%d][dir]" % i)

        i += 1

    return table_args


@blueprint.errorhandler(404)
def page_not_found(error):
    return render_template("error/404.html")


@blueprint.errorhandler(MasterDBNotFoundError)
def master_database_missing(error):
    log.exception(error)
    return render_template("error/master_database_missing.html")


@blueprint.errorhandler(StampedeDBNotFoundError)
def stampede_database_missing(error):
    log.exception(error)
    return render_template("error/stampede_database_missing.html")


@blueprint.errorhandler(DBAdminError)
def database_migration_error(error):
    log.exception(error)
    return render_template("error/database_migration_error.html", e=error)


@blueprint.errorhandler(ServiceError)
def error_response(error):
    log.exception(error)
    if request.is_xhr:
        return (
            serialize({"code": error.message.code, "message": error.message.message}),
            400,
            {"Content-Type": "application/json"},
        )
    else:
        return render_template("error/error_response.html", error=error.message)


@blueprint.errorhandler(Exception)
def catch_all(error):
    log.exception(error)
    return render_template("error/catch_all.html")



"""
Library to generate statistics from the new Stampede 3.1 backend.

Usage::
 stats = StampedeStatistics(connString='sqlite:///montage.db')
 stats.initialize('unique_wf_uuid')
 stats.set_job_filter('dax')
 print stats.get_total_jobs_status()
 print stats.get_total_succeeded_jobs_status()
 stats.set_job_filter('dag')
 print stats.get_total_jobs_status()
 print stats.get_total_succeeded_jobs_status()
 etc.
 stats.close()

Constructor and initialize methods:

The constructor takes a required sqlalchemy connection string
as the first argument.  The stats class will default to returning
data in the "expanded workflow" mode.  To change this behavior
and only analyize a single workflow set the optional arg:

expand_workflow = False

along with the connection string argument.

The initialize method is called with a single argument - the wf_uuid
of the desired "root workflow" whether returning data in expanded
mode or not.  The method will return True or False if a query
exception is raised so the programmer can test for success before
calling the subsequent query methods.  This method is intended
to be called once per object.

Job filtering:

Jobs can be filtered using any of the strings in the jobtype ENUM,
with the addition of the values 'all' and 'nonsub' which will
return all jobs and non-subworkflow jobs respectively.  If the
filter is not explicitly set, it will default to the 'all' mode.

The desired filter can be set with the set_job_filter() method. After
setting this method, all subsequent calls to the query methods will
return results according to the filter.  This can be set and reset
as many times as the user desires.  There is an example of re/setting
the job filter in the usage section above.  The query methods
will return different values after the filter is re/set.

Time filtering:

This behaves much like job filtering.  For the runtime queries,
the time intervals 'month', 'week', 'day', and  'hour' can
be set using the set_time_filter() method.  If this method
is not set, it will default to the 'month' interval for filtering.

Hostname filtering:

For the runtime queries the method set_host_filter() can be used to
filter by various hosts.  This method differs from the job and time
filtering methods in that the argument can be either a string (for
a single hostname), or an array/list of hostnames for multiple
hostnames.

Example::
 s.set_host_filter('butterfly.isi.edu')
 or
 s.set_host_filter(['engage-submit3.renci.org', 'node0012.palmetto.clemson.edu'])

Either one of these variations will work.  The first variation will
only retrieve data for that one host, the second will return data
for both hosts.  If this method is not set, no hostname filtering
will be done and information for all hosts will be returned.

Transformation filtering:

Transformation filtering works similarly to hostname filtering in
that it can accept a single string value or a array/list of strings.
However the set_transformation_filter() method accepts two keyword
arguments - 'include' and 'exclude'.  Only one of these keywords can
be set per method call.

Example::
 s.set_transformation_filter(include='pegasus::dirmanager')
 s.set_transformation_filter(exclude=['dagman::post' , 'dagman::pre' ,'condor::dagman'])
 etc.

This example demonstrates the two proper keyword invocations and
that either a string or list may be used.  If this method is not
set, no filtering will be done and information for all transforms
will be returned.  Calling this method with no arguments will
reset any previously set filters.

Return values from methods:

The return value types will vary from method to method.  Most of
the methods will return a single integer or floating point number.

Methods which return rows from the DB (rather than just a number)
will return a list which can be interacted with in one of two
ways - either by array index (list of tuples) or by a named attr
(list of objects).  The two following methods of interacting with
the same query results will both produce the same output:

Example::
 for row in s.get_job_kickstart():
     print row[0], row[1], row[2]
     print row.job_id, row.job_name, row.kickstart

Either syntax will work.  When using the named attribute method, the
attributes are the names of the columns/aliases in the SELECT
stanza of the query.  If the row returned by the method is printed,
it will display as a tuple of results per row.

Methods::
 get_sub_workflow_ids
 get_descendant_workflow_ids
 get_schema_version
 get_total_jobs_status
 get_total_succeeded_failed_jobs_status
 get_total_succeeded_jobs_status
 get_total_failed_jobs_status
 get_total_jobs_retries
 get_total_tasks_status
 get_total_succeeded_tasks_status
 get_total_failed_tasks_status
 get_task_success_report
 get_task_failure_report
 get_total_tasks_retries
 get_workflow_states
 get_workflow_cum_job_wall_time
 get_submit_side_job_wall_time
 get_workflow_details
 get_workflow_retries
 get_job_statistics
 get_job_states
 get_job_instance_sub_wf_map
 get_failed_job_instances
 get_job_instance_info
 get_job_name
 get_job_site
 get_job_kickstart
 get_job_runtime
 get_job_seqexec
 get_condor_q_time
 get_resource_delay
 get_post_time
 get_transformation_statistics
 get_invocation_by_time
 get_jobs_run_by_time
 get_invocation_by_time_per_host
 get_jobs_run_by_time_per_host

Methods listed in order of query list on wiki.

https://confluence.pegasus.isi.edu/display/pegasus/Pegasus+Statistics+Python+Version+Modified
"""
__author__ = "Monte Goode"

import logging

from sqlalchemy import orm
from sqlalchemy.sql.expression import and_, case, cast, distinct, func, not_, or_
from sqlalchemy.types import Float, Integer

from Pegasus.db import connection
from Pegasus.db.errors import StampedeDBNotFoundError
from Pegasus.db.schema import *
from Pegasus.service.dashboard import queries

# Main stats class.


class StampedeStatistics:
    def __init__(self, connString, expand_workflow=True):
        self.log = logging.getLogger(
            "{}.{}".format(self.__module__, self.__class__.__name__)
        )
        try:
            self.session = connection.connect(connString)
        except connection.ConnectionError as e:
            self.log.exception(e)
            raise StampedeDBNotFoundError

        self._expand = expand_workflow

        self._root_wf_id = None
        self._root_wf_uuid = None
        self._job_filter_mode = None
        self._time_filter_mode = None
        self._host_filter = None
        self._xform_filter = {"include": None, "exclude": None}
        self._connection_string = connString
        self._wfs = []

    def initialize(self, root_wf_uuid=None, root_wf_id=None):
        if root_wf_uuid is None and root_wf_id is None:
            self.log.error("Either root_wf_uuid or root_wf_id is required")
            raise ValueError("Either root_wf_uuid or root_wf_id is required")

        q = self.session.query(Workflow.root_wf_id, Workflow.wf_id, Workflow.wf_uuid)

        if root_wf_uuid:
            q = q.filter(Workflow.wf_uuid == root_wf_uuid)
        else:
            q = q.filter(Workflow.wf_id == root_wf_id)

        try:
            result = q.one()
            self._root_wf_id = result.wf_id
            self._root_wf_uuid = result.wf_uuid
            self._is_root_wf = result.root_wf_id == result.wf_id
        except orm.exc.MultipleResultsFound as e:
            self.log.error("Multiple results found for wf_uuid: %s", root_wf_uuid)
            raise
        except orm.exc.NoResultFound as e:
            self.log.error("No results found for wf_uuid: %s", root_wf_uuid)
            raise

        self._wfs.insert(0, self._root_wf_id)

        if self._expand:
            """
            select parent_wf_id, wf_id from workflow where root_wf_id =
            (select root_wf_id from workflow where wf_id=self._root_wf_id);
            """
            sub_q = (
                self.session.query(Workflow.root_wf_id)
                .filter(Workflow.wf_id == self._root_wf_id)
                .subquery("root_wf")
            )

            q = self.session.query(Workflow.parent_wf_id, Workflow.wf_id).filter(
                Workflow.root_wf_id == sub_q.c.root_wf_id
            )

            # @tree will hold the entire sub-work-flow dependency structure.
            tree = {}

            for row in q.all():
                parent_node = row.parent_wf_id
                if parent_node in tree:
                    tree[parent_node].append(row.wf_id)
                else:
                    tree[parent_node] = [row.wf_id]

            self._get_descendants(tree, self._root_wf_id, self._wfs)

        self.log.debug("Descendant workflow ids %s", self._wfs)

        if not len(self._wfs):
            self.log.error("No results found for wf_uuid: %s", root_wf_uuid)
            raise ValueError("No results found for wf_uuid: %s", root_wf_uuid)

        # Initialize filters with default value
        self.set_job_filter()
        self.set_time_filter()
        self.set_host_filter()
        self.set_transformation_filter()
        return True

    def get_descendants(self, wf_uuid):
        """
        Returns descendants of any workflow in a hierarchical workflow
        :param wf_uuid:
        :return: list of the workflow ids that are descendants including
                 the id for the node passed.
        """
        q = self.session.query(Workflow.root_wf_id, Workflow.wf_id, Workflow.wf_uuid)
        q = q.filter(Workflow.wf_uuid == wf_uuid)
        root_wf_id = None
        wf_id = None
        try:
            result = q.one()
            root_wf_id = result.root_wf_id
            wf_id = result.wf_id
        except orm.exc.MultipleResultsFound as e:
            self.log.error("Multiple results found for wf_uuid: %s", wf_uuid)
            raise
        except orm.exc.NoResultFound as e:
            self.log.error("No results found for wf_uuid: %s", wf_uuid)
            raise

        # will contain wf id's of the descendants of wf_uuid
        wfs = []
        wfs.insert(0, wf_id)

        """
        select parent_wf_id, wf_id from workflow where root_wf_id =
        (select root_wf_id from workflow where wf_id=root_wf_id);
        """
        sub_q = (
            self.session.query(Workflow.root_wf_id)
            .filter(Workflow.wf_id == root_wf_id)
            .subquery("root_wf")
        )

        q = self.session.query(Workflow.parent_wf_id, Workflow.wf_id).filter(
            Workflow.root_wf_id == sub_q.c.root_wf_id
        )

        # @tree will hold the entire sub-work-flow dependency structure
        # from the real root
        tree = {}

        for row in q.all():
            parent_node = row.parent_wf_id
            if parent_node in tree:
                tree[parent_node].append(row.wf_id)
            else:
                tree[parent_node] = [row.wf_id]

        self._get_descendants(tree, wf_id, wfs)

        self.log.debug("Descendant workflow ids %s", wfs)

        if not len(wfs):
            self.log.error("No results found for wf_uuid: %s", wf_uuid)
            raise ValueError("No results found for wf_uuid: %s", wf_uuid)
        return wfs

    def _get_descendants(self, tree, wf_node, wfs):
        """
        If the root_wf_uuid given to initialize function is not the UUID of the root work-flow, and
        expand_workflow was set to True, then this recursive function determines all child work-flows.
        :param tree: A dictionary when key is the parent_wf_id and value is a list of its child wf_id's.
        :param wf_node: The node for which to determine descendants.
        :param wfs: the list containing the descendants of the node.
        :return:
        """
        if tree is None or wf_node is None:
            raise ValueError("Tree, or node cannot be None")

        if wfs is None:
            raise ValueError(
                "wfs should be a list in which list of descendants has to be stored"
            )

        if wf_node in tree:

            wfs.extend(tree[wf_node])

            for wf in tree[wf_node]:
                self._get_descendants(tree, wf, wfs)

    def close(self):
        self.log.debug("close")
        self.session.close()

    def set_job_filter(self, filter="all"):
        modes = [
            "all",
            "nonsub",
            "subwf",
            "dax",
            "dag",
            "compute",
            "stage-in-tx",
            "stage-out-tx",
            "registration",
            "inter-site-tx",
            "create-dir",
            "staged-compute",
            "cleanup",
            "chmod",
        ]
        try:
            modes.index(filter)
            self._job_filter_mode = filter
            self.log.debug("Setting filter to: %s", filter)
        except Exception:
            self._job_filter_mode = "all"
            self.log.error("Unknown job filter %s - setting to all", filter)

    def set_time_filter(self, filter="month"):
        modes = ["month", "week", "day", "hour"]
        try:
            modes.index(filter)
            self._time_filter_mode = filter
            self.log.debug("Setting filter to: %s", filter)
        except Exception:
            self._time_filter_mode = "month"
            self.log.error("Unknown time filter %s - setting to month", filter)

    def set_host_filter(self, host=None):
        """
        The host argument can either be a string/single hostname or
        it can be a list/array of hostnames.
        """
        self._host_filter = host

    def set_transformation_filter(self, include=None, exclude=None):
        """
        Either of these args can either be a single string/xform type or
        it can be a list/array of xform types.

        Both arguments can not be set at the same time.  If they are,
        the program will log an error and not do any filtering.
        """
        self._xform_filter["include"] = include
        self._xform_filter["exclude"] = exclude

    #
    # Pulls information about sub workflows
    #

    def get_sub_workflow_ids(self):
        """
        Returns info on child workflows only.
        """
        q = self.session.query(Workflow.wf_id, Workflow.wf_uuid, Workflow.dax_label)
        q = q.filter(Workflow.parent_wf_id == self._root_wf_id)
        return q.all()

    def get_descendant_workflow_ids(self):
        q = self.session.query(Workflow.wf_id, Workflow.wf_uuid)
        q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        q = q.filter(Workflow.wf_id != self._root_wf_id)
        return q.all()

    def get_schema_version(self):
        return self.s_check.check_version()

    #
    # Status of initially planned wf components.
    #

    #
    # The following block of queries are documented here:
    # https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary
    # and
    # https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file
    #

    def _dax_or_dag_cond(self, JobO=Job):
        return or_(JobO.type_desc == "dax", JobO.type_desc == "dag")

    def _get_job_filter(self, JobO=Job):
        filters = {
            "all": None,
            "nonsub": not_(self._dax_or_dag_cond(JobO)),
            "subwf": self._dax_or_dag_cond(JobO),
            "dax": JobO.type_desc == "dax",
            "dag": JobO.type_desc == "dag",
            "compute": JobO.type_desc == "compute",
            "stage-in-tx": JobO.type_desc == "stage-in-tx",
            "stage-out-tx": JobO.type_desc == "stage-out-tx",
            "registration": JobO.type_desc == "registration",
            "inter-site-tx": JobO.type_desc == "inter-site-tx",
            "create-dir": JobO.type_desc == "create-dir",
            "staged-compute": JobO.type_desc == "staged-compute",
            "cleanup": JobO.type_desc == "cleanup",
            "chmod": JobO.type_desc == "chmod",
        }
        return filters[self._job_filter_mode]

    def _max_job_seq_subquery(self):
        """
        Creates the following subquery that is used in
        several queries:
        and jb_inst.job_submit_seq  = (
            select max(job_submit_seq) from job_instance where job_id = jb_inst.job_id group by job_id
            )
        """
        JobInstanceSubMax = orm.aliased(JobInstance)
        sub_q = self.session.query(
            func.max(JobInstanceSubMax.job_submit_seq).label("max_id")
        )
        sub_q = sub_q.filter(JobInstanceSubMax.job_id == JobInstance.job_id).correlate(
            JobInstance
        )
        sub_q = sub_q.group_by(JobInstanceSubMax.job_id).subquery()
        return sub_q

    def get_total_jobs_status(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totaljobs
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totaljobs
        """
        q = self.session.query(Job.job_id)
        if self._expand and self._is_root_wf:
            q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            q = q.filter(Workflow.wf_id.in_(self._wfs))
        else:
            q = q.filter(Workflow.wf_id == self._wfs[0])
        q = q.filter(Job.wf_id == Workflow.wf_id)
        if self._get_job_filter() is not None:
            q = q.filter(self._get_job_filter())

        return q.count()

    def get_total_succeeded_failed_jobs_status(self, classify_error=False, tag=None):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totalsucceeded_failed_jobs
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totalsucceededfailedjobs
        """
        JobInstanceSub = orm.aliased(JobInstance, name="JobInstanceSub")
        sq_1 = self.session.query(
            func.max(JobInstanceSub.job_submit_seq).label("jss"),
            JobInstanceSub.job_id.label("jobid"),
        )

        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(Workflow.wf_id == self._wfs[0])
        sq_1 = sq_1.filter(Workflow.wf_id == Job.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstanceSub.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())
        sq_1 = sq_1.group_by(JobInstanceSub.job_id).subquery()

        q = self.session.query(
            func.sum(case([(JobInstance.exitcode == 0, 1)], else_=0)).label(
                "succeeded"
            ),
            func.sum(case([(JobInstance.exitcode != 0, 1)], else_=0)).label("failed"),
        )
        q = q.filter(JobInstance.job_id == sq_1.c.jobid)
        q = q.filter(JobInstance.job_submit_seq == sq_1.c.jss)

        if classify_error:
            if tag is None:
                self.log.error("for error classification you need to specify tag")
                return None

            q = q.filter(JobInstance.job_instance_id == Tag.job_instance_id)
            q = q.filter(Tag.name == tag)
            q = q.filter(Tag.count > 0)

        return q.one()

    def get_total_held_jobs(self):
        """
        SELECT DISTINCT count(  job_instance_id) FROM
                                jobstate j JOIN   ( SELECT max(job_instance_id) as maxid FROM job_instance GROUP BY job_id) max_ji ON j.job_instance_id=max_ji.maxid
                                WHERE j.state = 'JOB_HELD';
        """

        sq_1 = self.session.query(
            func.max(JobInstance.job_instance_id).label("max_ji_id"),
            JobInstance.job_id.label("jobid"),
            Job.exec_job_id.label("jobname"),
        )

        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(Workflow.wf_id == self._wfs[0])

        sq_1 = sq_1.filter(Workflow.wf_id == Job.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstance.job_id)
        sq_1 = sq_1.group_by(JobInstance.job_id).subquery()

        q = self.session.query(
            distinct(Jobstate.job_instance_id.label("last_job_instance")),
            sq_1.c.jobid,
            sq_1.c.jobname,
            Jobstate.reason,
        )

        q = q.filter(Jobstate.state == "JOB_HELD")
        q = q.join(sq_1, Jobstate.job_instance_id == sq_1.c.max_ji_id)

        return q.all()

    def get_total_succeeded_jobs_status(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totalsucceededjobs
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totalsucceededjobs
        """
        JobInstanceSub = orm.aliased(JobInstance, name="JobInstanceSub")
        sq_1 = self.session.query(
            func.max(JobInstanceSub.job_submit_seq).label("jss"),
            JobInstanceSub.job_id.label("jobid"),
        )
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(Workflow.wf_id == self._wfs[0])
        sq_1 = sq_1.filter(Workflow.wf_id == Job.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstanceSub.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())
        sq_1 = sq_1.group_by(JobInstanceSub.job_id).subquery()

        q = self.session.query(JobInstance.job_instance_id.label("last_job_instance"))
        q = q.filter(JobInstance.job_id == sq_1.c.jobid)
        q = q.filter(JobInstance.job_submit_seq == sq_1.c.jss)
        q = q.filter(JobInstance.exitcode == 0).filter(
            JobInstance.exitcode != None
        )  # noqa: E711
        return q.count()

    def get_failing_jobs(self):
        """
        Returns jobs that are failing as in the most recent instance is running (exitcode is None).
        However, previous job retries have failed.
        :return:
        """

        # sanity check. this query does not expand to sub workflows. only the current
        # level of the workflow
        if self._expand:
            self.log.error(
                "For failing jobs expansion to sub workflows is not possible"
            )
            raise ValueError(
                "For failing jobs expansion to sub workflows is not possible"
            )

        try:
            workflow = queries.WorkflowInfo(
                self._connection_string,
                wf_id=self._root_wf_id,
                wf_uuid=self._root_wf_uuid,
            )
            return workflow.get_failing_jobs()
        finally:
            if workflow:
                workflow.close()

    def _get_total_failed_jobs_status(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totalfailedjobs
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totalfailedjobs
        """
        JobInstanceSub = orm.aliased(JobInstance, name="JobInstanceSub")
        sq_1 = self.session.query(
            func.max(JobInstanceSub.job_submit_seq).label("jss"),
            JobInstanceSub.job_id.label("jobid"),
        )
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(Workflow.wf_id == self._wfs[0])
        sq_1 = sq_1.filter(Workflow.wf_id == Job.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstanceSub.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())
        sq_1 = sq_1.group_by(JobInstanceSub.job_id).subquery()

        q = self.session.query(JobInstance.job_instance_id.label("last_job_instance"))
        q = q.filter(JobInstance.job_id == sq_1.c.jobid)
        q = q.filter(JobInstance.job_submit_seq == sq_1.c.jss)
        q = q.filter(JobInstance.exitcode != 0).filter(
            JobInstance.exitcode != None
        )  # noqa: E711

        return q

    def get_total_running_jobs_status(self):
        JobInstanceSub = orm.aliased(JobInstance, name="JobInstanceSub")
        sq_1 = self.session.query(
            func.max(JobInstanceSub.job_submit_seq).label("jss"),
            JobInstanceSub.job_id.label("jobid"),
        )
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(Workflow.wf_id == self._wfs[0])
        sq_1 = sq_1.filter(Workflow.wf_id == Job.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstanceSub.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())
        sq_1 = sq_1.group_by(JobInstanceSub.job_id).subquery()

        q = self.session.query(JobInstance.job_instance_id.label("last_job_instance"))
        q = q.filter(JobInstance.job_id == sq_1.c.jobid)
        q = q.filter(JobInstance.job_submit_seq == sq_1.c.jss)
        q = q.filter(JobInstance.exitcode == None)  # noqa: E711
        return q.count()

    def get_total_failed_jobs_status(self):

        q = self._get_total_failed_jobs_status()
        return q.count()

    def _query_jobstate_for_instance(self, states):
        """
        The states arg is a list of strings.
        Returns an appropriate subquery.
        """
        q = self.session.query(Jobstate.job_instance_id)
        q = q.filter(Jobstate.job_instance_id == JobInstance.job_instance_id).correlate(
            JobInstance
        )
        q = q.filter(Jobstate.state.in_(states)).subquery()
        return q

    def get_total_jobs_retries(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-TotalJobRetries
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-TotalJobRetries
        """
        self._dax_or_dag_cond()

        sq_1 = self.session.query(func.count(Job.job_id))
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(Workflow.wf_id == self._wfs[0])
        sq_1 = sq_1.filter(Job.wf_id == Workflow.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstance.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())

        sq_1 = sq_1.subquery()

        sq_2 = self.session.query(func.count(distinct(JobInstance.job_id)))
        if self._expand and self._is_root_wf:
            sq_2 = sq_2.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_2 = sq_2.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_2 = sq_2.filter(Workflow.wf_id == self._wfs[0])
        sq_2 = sq_2.filter(Job.wf_id == Workflow.wf_id)
        sq_2 = sq_2.filter(Job.job_id == JobInstance.job_id)
        if self._get_job_filter() is not None:
            sq_2 = sq_2.filter(self._get_job_filter())

        sq_2 = sq_2.subquery()

        q = self.session.query(
            (sq_1.as_scalar() - sq_2.as_scalar()).label("total_job_retries")
        )

        return q.all()[0].total_job_retries

    def get_total_tasks_status(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totaltask
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totaltasks
        """
        q = self.session.query(Task.task_id)
        if self._expand and self._is_root_wf:
            q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            q = q.filter(Workflow.wf_id.in_(self._wfs))
        else:
            q = q.filter(Workflow.wf_id == self._wfs[0])
        q = q.filter(Task.wf_id == Workflow.wf_id)
        q = q.filter(Task.job_id == Job.job_id)
        if self._get_job_filter(Task) is not None:
            q = q.filter(self._get_job_filter(Task))
        return q.count()

    def _base_task_status_query_old(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totalsucceededtasks
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totalsucceededtasks
        """
        # This query generation method is obsolete and is only being
        # kept for optimization reference.

        WorkflowSub1 = orm.aliased(Workflow, name="WorkflowSub1")
        JobInstanceSub1 = orm.aliased(JobInstance, name="JobInstanceSub1")
        JobSub1 = orm.aliased(Job, name="JobSub1")

        sq_1 = self.session.query(
            WorkflowSub1.wf_id.label("wid"),
            func.max(JobInstanceSub1.job_submit_seq).label("jss"),
            JobInstanceSub1.job_id.label("jobid"),
        )
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(WorkflowSub1.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(WorkflowSub1.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(WorkflowSub1.wf_id == self._wfs[0])
        sq_1 = sq_1.filter(WorkflowSub1.wf_id == JobSub1.wf_id)
        sq_1 = sq_1.filter(JobSub1.job_id == JobInstanceSub1.job_id)
        sq_1 = sq_1.group_by(JobInstanceSub1.job_id)
        if self._get_job_filter(JobSub1) is not None:
            sq_1 = sq_1.filter(self._get_job_filter(JobSub1))
        sq_1 = sq_1.subquery()

        JobInstanceSub2 = orm.aliased(JobInstance, name="JobInstanceSub2")
        sq_2 = self.session.query(
            sq_1.c.wid.label("wf_id"),
            JobInstanceSub2.job_instance_id.label("last_job_instance_id"),
        )
        sq_2 = sq_2.filter(JobInstanceSub2.job_id == sq_1.c.jobid)
        sq_2 = sq_2.filter(JobInstanceSub2.job_submit_seq == sq_1.c.jss)
        sq_2 = sq_2.subquery()

        q = self.session.query(Invocation.invocation_id)
        q = q.filter(Invocation.abs_task_id != None)  # noqa: E711
        q = q.filter(Invocation.job_instance_id == sq_2.c.last_job_instance_id)
        q = q.filter(Invocation.wf_id == sq_2.c.wf_id)

        # Calling wrapper methods would invoke like so:
        # q = self._base_task_status_query()
        # q = q.filter(Invocation.exitcode == 0)
        # return q.count()

        return q

    def _base_task_statistics_query(self, success=True, pmc=False):
        w = orm.aliased(Workflow, name="w")
        j = orm.aliased(Job, name="j")
        ji = orm.aliased(JobInstance, name="ji")

        sq_1 = self.session.query(
            w.wf_id,
            j.job_id,
            ji.job_instance_id.label("jiid"),
            ji.job_submit_seq.label("jss"),
            func.max(ji.job_submit_seq).label("maxjss"),
        )
        if pmc:
            sq_1 = self.session.query(
                w.wf_id,
                j.job_id,
                ji.job_instance_id.label("jiid"),
                ji.job_submit_seq.label("jss"),
            )

        sq_1 = sq_1.join(j, w.wf_id == j.wf_id)
        sq_1 = sq_1.join(ji, j.job_id == ji.job_id)
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(w.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(w.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(w.wf_id == self._wfs[0])
        if not pmc:
            sq_1 = sq_1.group_by(j.job_id)
        if self._get_job_filter(j) is not None:
            sq_1 = sq_1.filter(self._get_job_filter(j))
        sq_1 = sq_1.subquery("t")

        # PM-713 - Change to func.count(distinct(Invocation.abs_task_id)) from func.count(Invocation.exitcode)
        sq_2 = self.session.query(
            sq_1.c.wf_id, func.count(distinct(Invocation.abs_task_id)).label("count")
        )
        sq_2 = sq_2.select_from(
            orm.join(sq_1, Invocation, sq_1.c.jiid == Invocation.job_instance_id)
        )
        if not pmc:
            sq_2 = sq_2.filter(sq_1.c.jss == sq_1.c.maxjss)

        sq_2 = sq_2.filter(Invocation.abs_task_id != None)  # noqa: E711
        if success:
            sq_2 = sq_2.filter(Invocation.exitcode == 0)
        else:
            sq_2 = sq_2.filter(Invocation.exitcode != 0)
        sq_2 = sq_2.group_by(sq_1.c.wf_id)
        return sq_2

    def _task_statistics_query_sum(self, success=True, pmc=False):
        s = self._base_task_statistics_query(success, pmc).subquery("tt")
        q = self.session.query(func.sum(s.c.count).label("task_count"))
        return q.one()[0] or 0

    def get_total_succeeded_tasks_status(self, pmc=False):
        return self._task_statistics_query_sum(True, pmc)

    def get_total_failed_tasks_status(self):
        return self._task_statistics_query_sum(False, False)

    def get_task_success_report(self, pmc=False):
        return self._base_task_statistics_query(True, pmc).all()

    def get_task_failure_report(self):
        return self._base_task_statistics_query(False, False).all()

    def get_total_tasks_retries(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Totaltaskretries
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Totaltaskretries
        """
        sq_1 = self.session.query(
            Workflow.wf_id.label("wid"), Invocation.abs_task_id.label("tid")
        )
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        elif self._expand and not self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        else:
            sq_1 = sq_1.filter(Workflow.wf_id == self._wfs[0])
        sq_1 = sq_1.filter(Job.wf_id == Workflow.wf_id)
        sq_1 = sq_1.filter(Invocation.wf_id == Workflow.wf_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstance.job_id)
        if self._get_job_filter() is not None:
            sq_1 = sq_1.filter(self._get_job_filter())
        sq_1 = sq_1.filter(JobInstance.job_instance_id == Invocation.job_instance_id)
        sq_1 = sq_1.filter(Invocation.abs_task_id != None)  # noqa: E711

        i = 0
        f = {}
        for row in sq_1.all():
            i += 1
            if row not in f:
                f[row] = True

        return i - len(f.keys())

    #
    # Run statistics
    #

    def get_workflow_states(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Workflowwalltime
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowwalltime
        """
        q = self.session.query(
            Workflowstate.wf_id,
            Workflowstate.state,
            Workflowstate.timestamp,
            Workflowstate.restart_count,
            Workflowstate.status,
        )
        q = q.filter(Workflowstate.wf_id == self._root_wf_id).order_by(
            Workflowstate.restart_count
        )

        return q.all()

    def get_workflow_cum_job_wall_time(self):
        """
        select sum(remote_duration * multiplier_factor) FROM
        invocation as invoc, job_instance as ji WHERE
        invoc.task_submit_seq >= 0 and
        invoc.job_instance_id = ji.job_instance_id and
        invoc.wf_id in (1,2,3) and
        invoc.transformation <> 'condor::dagman'

        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Workflowcumulativejobwalltime
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowcumulativejobwalltime
        """
        q = self.session.query(
            cast(
                func.sum(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ),
            cast(
                func.sum(
                    case(
                        [
                            (
                                Invocation.exitcode == 0,
                                Invocation.remote_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("goodput"),
                Float,
            ),
            cast(
                func.sum(
                    case(
                        [
                            (
                                Invocation.exitcode > 0,
                                Invocation.remote_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("badput"),
                Float,
            ),
        )

        q = q.filter(Invocation.task_submit_seq >= 0)
        q = q.filter(Invocation.job_instance_id == JobInstance.job_instance_id)

        if self._expand:
            q = q.filter(Invocation.wf_id == Workflow.wf_id)
            q = q.filter(Workflow.root_wf_id == self._root_wf_id)

        else:
            q = q.filter(Invocation.wf_id.in_(self._wfs))

        q = q.filter(Invocation.transformation != "condor::dagman")

        return q.first()

    def get_summary_integrity_metrics(self):
        """

        :param type:    whether integrity type is check | compute
        :param file_type: file type input or output
        :return:
        """
        q = self.session.query(
            IntegrityMetrics.type,
            func.sum(IntegrityMetrics.duration).label("duration"),
            func.sum(IntegrityMetrics.count).label("count"),
        )

        q = q.group_by(IntegrityMetrics.type)

        if self._expand:
            q = q.filter(IntegrityMetrics.wf_id == Workflow.wf_id)
            q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        else:
            q = q.filter(IntegrityMetrics.wf_id.in_(self._wfs))

        # at most two records grouped by type compute | check
        return q.all()

    def get_tag_metrics(self, name):
        """

        :param name:    what type of tag to aggregate on
        :return:
        """
        q = self.session.query(Tag.name, func.sum(Tag.count).label("count"))

        q = q.group_by(Tag.name)
        q = q.filter(Tag.name == name)

        if self._expand:
            q = q.filter(Tag.wf_id == Workflow.wf_id)
            q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        else:
            q = q.filter(Tag.wf_id.in_(self._wfs))

        return q.all()

    def get_integrity_metrics(self):
        """

        :param type:    whether integrity type is check | compute
        :param file_type: file type input or output
        :return:
        """
        q = self.session.query(
            IntegrityMetrics.type,
            IntegrityMetrics.file_type,
            func.sum(IntegrityMetrics.duration).label("duration"),
            func.sum(IntegrityMetrics.count).label("count"),
        )

        q = q.group_by(IntegrityMetrics.type)
        q = q.group_by(IntegrityMetrics.file_type)

        if self._expand:
            q = q.filter(IntegrityMetrics.wf_id == Workflow.wf_id)
            q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        else:
            q = q.filter(IntegrityMetrics.wf_id.in_(self._wfs))

        """
        for result in q.all():
            print result
            print result.type
            print result.file_type
        """
        return q.all()

    def get_submit_side_job_wall_time(self):
        """
        select sum(local_duration * multiplier_factor) FROM
        job_instance as jb_inst, job as jb WHERE
        jb_inst.job_id  = jb.job_id and
        jb.wf_id in (1,2,3) and
        ((not (jb.type_desc ='dax' or jb.type_desc ='dag'))
          or
         ((jb.type_desc ='dax' or jb.type_desc ='dag') and jb_inst.subwf_id is NULL)
        )

        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Summary#WorkflowSummary-Cumulativejobwalltimeasseenfromsubmitside
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Cumulativejobwalltimeasseenfromsubmitside
        """
        q = self.session.query(
            cast(
                func.sum(JobInstance.local_duration * JobInstance.multiplier_factor),
                Float,
            ).label("wall_time"),
            cast(
                func.sum(
                    case(
                        [
                            (
                                JobInstance.exitcode == 0,
                                JobInstance.local_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("goodput"),
                Float,
            ),
            cast(
                func.sum(
                    case(
                        [
                            (
                                JobInstance.exitcode > 0,
                                JobInstance.local_duration
                                * JobInstance.multiplier_factor,
                            )
                        ],
                        else_=0,
                    )
                ).label("badput"),
                Float,
            ),
        )

        q = q.filter(JobInstance.job_id == Job.job_id)

        if self._expand:
            q = q.filter(Job.wf_id == Workflow.wf_id)
            q = q.filter(Workflow.root_wf_id == self._root_wf_id)

        else:
            q = q.filter(Job.wf_id.in_(self._wfs))

        if self._expand:
            d_or_d = self._dax_or_dag_cond()
            q = q.filter(
                or_(not_(d_or_d), and_(d_or_d, JobInstance.subwf_id == None))
            )  # noqa: E711

        return q.first()

    def get_workflow_details(self, wfs=None):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowdetails
        :param wfs:  list of workflow ids
        :return: workflow details for the workflows queried
        """

        if wfs is None:
            wfs = self._wfs

        q = self.session.query(
            Workflow.wf_id,
            Workflow.wf_uuid,
            Workflow.parent_wf_id,
            Workflow.root_wf_id,
            Workflow.dag_file_name,
            Workflow.submit_hostname,
            Workflow.submit_dir,
            Workflow.planner_arguments,
            Workflow.user,
            Workflow.grid_dn,
            Workflow.planner_version,
            Workflow.dax_label,
            Workflow.dax_version,
        )
        q = q.filter(Workflow.wf_id.in_(wfs))
        return q.all()

    def get_workflow_retries(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Workflow+Statistics+file#WorkflowStatisticsfile-Workflowretries
        """
        sq_1 = self.session.query(func.max(Workflowstate.restart_count).label("retry"))
        if self._expand and self._is_root_wf:
            sq_1 = sq_1.filter(Workflow.root_wf_id == self._root_wf_id)
        else:
            sq_1 = sq_1.filter(Workflow.wf_id.in_(self._wfs))
        sq_1 = sq_1.filter(Workflowstate.wf_id == Workflow.wf_id)
        sq_1 = sq_1.group_by(Workflowstate.wf_id)
        sq_1 = sq_1.subquery()

        q = self.session.query(func.sum(sq_1.c.retry).label("total_retry"))
        return q.one().total_retry

    #
    # Job Statistics
    # These queries are documented:
    # https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file
    #

    def get_job_statistics(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-All
        """
        if self._expand:
            return []
        sq_1 = self.session.query(func.min(Jobstate.timestamp))
        sq_1 = sq_1.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_1 = sq_1.filter(
            or_(
                Jobstate.state == "GRID_SUBMIT",
                Jobstate.state == "GLOBUS_SUBMIT",
                Jobstate.state == "EXECUTE",
            )
        )
        sq_1 = sq_1.subquery()

        sq_2 = self.session.query(Jobstate.timestamp)
        sq_2 = sq_2.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_2 = sq_2.filter(Jobstate.state == "SUBMIT")
        sq_2 = sq_2.subquery()

        sq_3 = self.session.query(func.min(Jobstate.timestamp))
        sq_3 = sq_3.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_3 = sq_3.filter(Jobstate.state == "EXECUTE")
        sq_3 = sq_3.subquery()

        sq_4 = self.session.query(func.min(Jobstate.timestamp))
        sq_4 = sq_4.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_4 = sq_4.filter(
            or_(Jobstate.state == "GRID_SUBMIT", Jobstate.state == "GLOBUS_SUBMIT")
        )
        sq_4 = sq_4.subquery()

        sq_5 = self.session.query(func.sum(Invocation.remote_duration))
        sq_5 = sq_5.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_5 = sq_5.filter(Invocation.wf_id == Job.wf_id).correlate(Job)
        sq_5 = sq_5.filter(Invocation.task_submit_seq >= 0)
        sq_5 = sq_5.group_by().subquery()

        sq_6 = self.session.query(Jobstate.timestamp)
        sq_6 = sq_6.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_6 = sq_6.filter(Jobstate.state == "POST_SCRIPT_TERMINATED")
        sq_6 = sq_6.subquery()

        sq_7 = self.session.query(func.max(Jobstate.timestamp))
        sq_7 = sq_7.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_7 = sq_7.filter(
            or_(
                Jobstate.state == "POST_SCRIPT_STARTED",
                Jobstate.state == "JOB_TERMINATED",
            )
        )
        sq_7 = sq_7.subquery()

        sq_8 = self.session.query(func.max(Invocation.exitcode))
        sq_8 = sq_8.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_8 = sq_8.filter(Invocation.wf_id == Job.wf_id).correlate(Job)
        # PM-704 the task submit sequence needs to be >= -1 to include prescript status
        sq_8 = sq_8.filter(Invocation.task_submit_seq >= -1)
        sq_8 = sq_8.group_by().subquery()

        JobInstanceSub = orm.aliased(JobInstance)

        sq_9 = self.session.query(Host.hostname)
        sq_9 = sq_9.filter(
            JobInstanceSub.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_9 = sq_9.filter(Host.host_id == JobInstanceSub.host_id)
        sq_9 = sq_9.subquery()

        JI = orm.aliased(JobInstance)
        sq_10 = self.session.query(
            func.sum(Invocation.remote_duration * JI.multiplier_factor)
        )
        sq_10 = sq_10.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_10 = sq_10.filter(Invocation.job_instance_id == JI.job_instance_id)
        sq_10 = sq_10.filter(Invocation.wf_id == Job.wf_id).correlate(Job)
        sq_10 = sq_10.filter(Invocation.task_submit_seq >= 0)
        sq_10 = sq_10.group_by().subquery()

        sq_11 = self.session.query(func.sum(Invocation.remote_cpu_time))
        sq_11 = sq_11.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_11 = sq_11.filter(Invocation.wf_id == Job.wf_id).correlate(Job)
        sq_11 = sq_11.filter(Invocation.task_submit_seq >= 0)
        sq_11 = sq_11.group_by().subquery()

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            Job.exec_job_id.label("job_name"),
            JobInstance.site,
            cast(sq_1.as_scalar() - sq_2.as_scalar(), Float).label("condor_q_time"),
            cast(sq_3.as_scalar() - sq_4.as_scalar(), Float).label("resource_delay"),
            cast(JobInstance.local_duration, Float).label("runtime"),
            cast(sq_5.as_scalar(), Float).label("kickstart"),
            cast(sq_6.as_scalar() - sq_7.as_scalar(), Float).label("post_time"),
            cast(JobInstance.cluster_duration, Float).label("seqexec"),
            sq_8.as_scalar().label("exit_code"),
            sq_9.as_scalar().label("host_name"),
            JobInstance.multiplier_factor,
            cast(sq_10.as_scalar(), Float).label("kickstart_multi"),
            sq_11.as_scalar().label("remote_cpu_time"),
        )
        q = q.filter(JobInstance.job_id == Job.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.order_by(JobInstance.job_submit_seq)

        return q.all()

    def _state_sub_q(self, states, function=None):
        sq = None
        if not function:
            sq = self.session.query(Jobstate.timestamp)
        elif function == "max":
            sq = self.session.query(func.max(Jobstate.timestamp))
        elif function == "min":
            sq = self.session.query(func.min(Jobstate.timestamp))
        sq = sq.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq = sq.filter(Jobstate.state.in_(states)).subquery()
        return sq

    def get_job_states(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-JobStates
        """
        if self._expand:
            return []
        sq_1 = (
            self.session.query(Host.hostname)
            .filter(Host.host_id == JobInstance.host_id)
            .correlate(JobInstance)
            .subquery()
        )
        # select min(timestamp) from jobstate where job_instance_id = jb_inst.job_instance_id
        # ) as jobS ,
        # (
        # select max(timestamp)-min(timestamp) from jobstate where job_instance_id = jb_inst.job_instance_id
        # ) as jobDuration,

        sq_jobS = self.session.query(func.min(Jobstate.timestamp))
        sq_jobS = (
            sq_jobS.filter(Jobstate.job_instance_id == JobInstance.job_instance_id)
            .correlate(JobInstance)
            .subquery()
        )

        sq_jobD = self.session.query(
            func.max(Jobstate.timestamp) - func.min(Jobstate.timestamp)
        )
        sq_jobD = (
            sq_jobD.filter(Jobstate.job_instance_id == JobInstance.job_instance_id)
            .correlate(JobInstance)
            .subquery()
        )

        sq_2 = self._state_sub_q(["PRE_SCRIPT_STARTED"])
        sq_3 = self._state_sub_q(["PRE_SCRIPT_TERMINATED"])
        sq_4 = self._state_sub_q(["PRE_SCRIPT_STARTED"])
        sq_5 = self._state_sub_q(["SUBMIT"])
        sq_6 = self._state_sub_q(["JOB_TERMINATED"])
        sq_7 = self._state_sub_q(["GRID_SUBMIT", "GLOBUS_SUBMIT"], "max")
        sq_8 = self._state_sub_q(["EXECUTE"], "min")
        sq_9 = self._state_sub_q(["EXECUTE", "SUBMIT"], "max")
        sq_10 = self._state_sub_q(["JOB_TERMINATED"])

        sq_11 = self.session.query(func.min(Invocation.start_time))
        sq_11 = sq_11.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_11 = sq_11.filter(Invocation.wf_id == Job.wf_id).correlate(Job)
        sq_11 = sq_11.filter(Invocation.task_submit_seq >= 0)
        sq_11 = sq_11.group_by(Invocation.job_instance_id).subquery()

        sq_12 = self.session.query(func.sum(Invocation.remote_duration))
        sq_12 = sq_12.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_12 = sq_12.filter(Invocation.wf_id == Job.wf_id).correlate(Job)
        sq_12 = sq_12.filter(Invocation.task_submit_seq >= 0)
        sq_12 = sq_12.group_by(Invocation.job_instance_id).subquery()

        sq_13 = self._state_sub_q(["POST_SCRIPT_STARTED", "JOB_TERMINATED"], "max")
        sq_14 = self._state_sub_q(["POST_SCRIPT_TERMINATED"])

        sq_15 = self.session.query(
            func.group_concat(func.distinct(Invocation.transformation))
        )
        sq_15 = sq_15.filter(Invocation.wf_id.in_(self._wfs))
        sq_15 = sq_15.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_15 = sq_15.filter(Invocation.transformation != "dagman::post")
        sq_15 = sq_15.filter(Invocation.transformation != "dagman::pre")
        sq_15 = sq_15.subquery()

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            Job.exec_job_id.label("job_name"),
            JobInstance.site,
            sq_1.as_scalar().label("host_name"),
            cast(sq_jobS.as_scalar(), Float).label("jobS"),
            cast(sq_jobD.as_scalar(), Float).label("jobDuration"),
            cast(sq_2.as_scalar(), Float).label("pre_start"),
            cast(sq_3.as_scalar() - sq_4.as_scalar(), Float).label("pre_duration"),
            cast(sq_5.as_scalar(), Float).label("condor_start"),
            cast(sq_6.as_scalar() - sq_5.as_scalar(), Float).label("condor_duration"),
            cast(sq_7.as_scalar(), Float).label("grid_start"),
            cast(sq_8.as_scalar() - sq_7.as_scalar(), Float).label("grid_duration"),
            cast(sq_9.as_scalar(), Float).label("exec_start"),
            cast(sq_10.as_scalar() - sq_9.as_scalar(), Float).label("exec_duration"),
            cast(sq_11.as_scalar(), Float).label("kickstart_start"),
            cast(sq_12.as_scalar(), Float).label("kickstart_duration"),
            cast(sq_13.as_scalar(), Float).label("post_start"),
            cast(sq_14.as_scalar() - sq_13.as_scalar(), Float).label("post_duration"),
            sq_15.as_scalar().label("transformation"),
        )
        q = q.filter(JobInstance.job_id == Job.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.order_by(JobInstance.job_submit_seq)

        return q.all()

    def get_job_instance_sub_wf_map(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Subworkflowjobinstancesmapping
        """
        if self._expand:
            return []
        q = self.session.query(JobInstance.job_instance_id, JobInstance.subwf_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(self._dax_or_dag_cond())
        return q.all()

    def get_failed_job_instances(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Failedjobinstances
        """

        # PM-752 we use the same query that we used to get the count of failed jobs
        q = self._get_total_failed_jobs_status()
        return q.all()

    def get_plots_failed_job_instances(self, final=False, all_jobs=False):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Failedjobinstances
        used in the pegasus plots code. is deprecated
        """
        if self._expand:
            return []
        d_or_d = self._dax_or_dag_cond()

        if not final:
            q = self.session.query(
                JobInstance.job_instance_id, JobInstance.job_submit_seq
            )
        else:
            q = self.session.query(
                JobInstance.job_instance_id, func.max(JobInstance.job_submit_seq)
            )
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.filter(Job.job_id == JobInstance.job_id)
        if not all_jobs:
            q = q.filter(
                or_(not_(d_or_d), and_(d_or_d, JobInstance.subwf_id == None))
            )  # noqa: E711
        q = q.filter(JobInstance.exitcode != 0).filter(
            JobInstance.exitcode != None
        )  # noqa: E711
        if final:
            q = q.group_by(JobInstance.job_id)
        q = q.order_by(JobInstance.job_submit_seq)

        return q.all()

    def get_job_instance_info(self, job_instance_id=None):
        """
        Job instance information.  Pulls all or for one instance.
        https://confluence.pegasus.isi.edu/pages/viewpage.action?pageId=14876831
        """
        if self._expand:
            return []

        sq_0 = self.session.query(Workflow.submit_dir)
        sq_0 = sq_0.filter(Workflow.wf_id == JobInstance.subwf_id).correlate(
            JobInstance
        )
        sq_0 = sq_0.subquery()

        sq_1 = self.session.query(Job.exec_job_id)
        sq_1 = sq_1.filter(Job.job_id == JobInstance.job_id).correlate(JobInstance)
        sq_1 = sq_1.subquery()

        sq_2 = self.session.query(Job.submit_file)
        sq_2 = sq_2.filter(Job.job_id == JobInstance.job_id).correlate(JobInstance)
        sq_2 = sq_2.subquery()

        sq_3 = self.session.query(Job.executable)
        sq_3 = sq_3.filter(Job.job_id == JobInstance.job_id).correlate(JobInstance)
        sq_3 = sq_3.subquery()

        sq_4 = self.session.query(Job.argv)
        sq_4 = sq_4.filter(Job.job_id == JobInstance.job_id).correlate(JobInstance)
        sq_4 = sq_4.subquery()

        sq_5 = self.session.query(Workflow.submit_dir)
        sq_5 = sq_5.filter(Workflow.wf_id == self._root_wf_id).subquery()

        sq_6 = self.session.query(
            func.max(Jobstate.jobstate_submit_seq).label("max_job_submit_seq")
        )
        sq_6 = sq_6.filter(Jobstate.job_instance_id == job_instance_id)
        sq_6 = sq_6.subquery()

        sq_7 = self.session.query(Jobstate.state)
        sq_7 = sq_7.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_7 = sq_7.filter(Jobstate.jobstate_submit_seq == sq_6.as_scalar())
        sq_7 = sq_7.subquery()

        sq_8 = self.session.query(Invocation.executable)
        sq_8 = sq_8.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_8 = sq_8.filter(Invocation.task_submit_seq == -1)
        sq_8 = sq_8.subquery()

        sq_9 = self.session.query(Invocation.argv)
        sq_9 = sq_9.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_9 = sq_9.filter(Invocation.task_submit_seq == -1)
        sq_9 = sq_9.subquery()

        sq_10 = self.session.query(Host.hostname)
        sq_10 = sq_10.filter(Host.host_id == JobInstance.host_id).correlate(JobInstance)
        sq_10 = sq_10.subquery()

        q = self.session.query(
            JobInstance.job_instance_id,
            JobInstance.site,
            JobInstance.stdout_file,
            JobInstance.stderr_file,
            JobInstance.stdout_text,
            JobInstance.stderr_text,
            JobInstance.work_dir,
            sq_0.as_scalar().label("subwf_dir"),
            sq_1.as_scalar().label("job_name"),
            sq_2.as_scalar().label("submit_file"),
            sq_3.as_scalar().label("executable"),
            sq_4.as_scalar().label("argv"),
            sq_5.as_scalar().label("submit_dir"),
            sq_7.as_scalar().label("state"),
            sq_8.as_scalar().label("pre_executable"),
            sq_9.as_scalar().label("pre_argv"),
            sq_10.as_scalar().label("hostname"),
        )

        if job_instance_id:
            q = q.filter(JobInstance.job_instance_id == job_instance_id)

        return q.all()

    def get_invocation_info(self, ji_id=None):
        """
        SELECT task_submit_seq, exitcode, executable, argv, transformation, abs_task_id
        FROM invocation WHERE job_instance_id = 7 and wf_id = 1
        """
        if self._expand or not ji_id:
            return []

        q = self.session.query(
            Invocation.task_submit_seq,
            Invocation.exitcode,
            Invocation.executable,
            Invocation.argv,
            Invocation.transformation,
            Invocation.abs_task_id,
        )
        q = q.filter(Invocation.job_instance_id == ji_id)
        q = q.filter(Invocation.wf_id.in_(self._wfs))

        return q.all()

    def get_job_name(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Name
        """
        if self._expand:
            return []
        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            Job.exec_job_id.label("job_name"),
        )
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs)).order_by(JobInstance.job_submit_seq)
        return q.all()

    def get_job_site(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Site
        """
        if self._expand:
            return []
        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            JobInstance.site,
        )
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.filter(Job.job_id == JobInstance.job_id).group_by(Job.job_id)
        q = q.order_by(JobInstance.job_submit_seq)
        return q.all()

    def get_job_kickstart(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Kickstart
        """
        if self._expand:
            return []

        sq_1 = self.session.query(
            func.sum(Invocation.remote_duration * JobInstance.multiplier_factor)
        )
        sq_1 = sq_1.filter(
            Invocation.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_1 = sq_1.filter(Invocation.wf_id == Job.wf_id).correlate(Job)
        sq_1 = sq_1.filter(Invocation.task_submit_seq >= 0)
        sq_1 = sq_1.group_by(Invocation.job_instance_id)
        sq_1 = sq_1.subquery()

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            cast(sq_1.as_scalar(), Float).label("kickstart"),
        )
        q = q.filter(JobInstance.job_id == Job.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.order_by(JobInstance.job_submit_seq)
        return q.all()

    def get_job_runtime(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Runtime
        """
        if self._expand:
            return []

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            JobInstance.local_duration.label("runtime"),
        )
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.group_by(Job.job_id).order_by(JobInstance.job_submit_seq)

        return q.all()

    def get_job_seqexec(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Seqexec
        """
        if self._expand:
            return []
        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            JobInstance.cluster_duration,
        )
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.filter(Job.clustered != 0)
        q = q.order_by(JobInstance.job_submit_seq)

        return q.all()

    def get_condor_q_time(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-CondorQTime
        """
        if self._expand:
            return []

        sq_1 = self.session.query(func.min(Jobstate.timestamp))
        sq_1 = sq_1.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_1 = sq_1.filter(
            or_(
                Jobstate.state == "GRID_SUBMIT",
                Jobstate.state == "GLOBUS_SUBMIT",
                Jobstate.state == "EXECUTE",
            )
        )
        sq_1 = sq_1.subquery()

        sq_2 = self.session.query(Jobstate.timestamp)
        sq_2 = sq_2.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_2 = sq_2.filter(Jobstate.state == "SUBMIT")
        sq_2 = sq_2.subquery()

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            cast(sq_1.as_scalar() - sq_2.as_scalar(), Float).label("condor_q_time"),
        )
        q = q.filter(JobInstance.job_id == Job.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.order_by(JobInstance.job_submit_seq)

        return q.all()

    def get_resource_delay(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Resource
        """
        if self._expand:
            return []

        sq_1 = self.session.query(func.min(Jobstate.timestamp))
        sq_1 = sq_1.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_1 = sq_1.filter(Jobstate.state == "EXECUTE")
        sq_1 = sq_1.subquery()

        sq_2 = self.session.query(Jobstate.timestamp)
        sq_2 = sq_2.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_2 = sq_2.filter(
            or_(Jobstate.state == "GRID_SUBMIT", Jobstate.state == "GLOBUS_SUBMIT")
        )
        sq_2 = sq_2.subquery()

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            cast(sq_1.as_scalar() - sq_2.as_scalar(), Float).label("resource_delay"),
        )
        q = q.filter(JobInstance.job_id == Job.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.order_by(JobInstance.job_submit_seq)

        return q.all()

    def get_post_time(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Job+Statistics+file#JobStatisticsfile-Post
        """
        if self._expand:
            return []

        sq_1 = self.session.query(Jobstate.timestamp)
        sq_1 = sq_1.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_1 = sq_1.filter(Jobstate.state == "POST_SCRIPT_TERMINATED")
        sq_1 = sq_1.subquery()

        sq_2 = self.session.query(func.max(Jobstate.timestamp))
        sq_2 = sq_2.filter(
            Jobstate.job_instance_id == JobInstance.job_instance_id
        ).correlate(JobInstance)
        sq_2 = sq_2.filter(
            or_(
                Jobstate.state == "POST_SCRIPT_STARTED",
                Jobstate.state == "JOB_TERMINATED",
            )
        )
        sq_2 = sq_2.subquery()

        q = self.session.query(
            Job.job_id,
            JobInstance.job_instance_id,
            JobInstance.job_submit_seq,
            cast(sq_1.as_scalar() - sq_2.as_scalar(), Float).label("post_time"),
        )
        q = q.filter(JobInstance.job_id == Job.job_id)
        q = q.filter(Job.wf_id.in_(self._wfs))
        q = q.order_by(JobInstance.job_submit_seq)

        return q.all()

    #
    # This query documented:
    # https://confluence.pegasus.isi.edu/display/pegasus/Transformation+Statistics+file
    #

    def get_transformation_statistics(self):
        """
        SELECT transformation,
               count(invocation_id) as count,
               min(remote_duration * multiplier_factor) as min,
               count(CASE WHEN (invoc.exitcode = 0 and invoc.exitcode is NOT NULL) THEN invoc.exitcode END) AS success,
               count(CASE WHEN (invoc.exitcode != 0 and invoc.exitcode is NOT NULL) THEN invoc.exitcode END) AS failure,
               max(remote_duration * multiplier_factor) as max,
               avg(remote_duration * multiplier_factor) as avg,
               sum(remote_duration * multiplier_factor) as sum
        FROM
        invocation as invoc, job_instance as ji WHERE
        invoc.job_instance_id = ji.job_instance_id and
        invoc.wf_id IN (1,2,3) GROUP BY transformation
        """
        q = self.session.query(
            Invocation.transformation,
            case([(Invocation.exitcode == 0, "successful")], else_="failed").label(
                "type"
            ),
            func.count(Invocation.invocation_id).label("count"),
            func.count(case([(Invocation.exitcode == 0, Invocation.exitcode)])).label(
                "success"
            ),
            func.count(case([(Invocation.exitcode != 0, Invocation.exitcode)])).label(
                "failure"
            ),
            # runtime
            cast(
                func.min(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("min"),
            cast(
                func.max(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("max"),
            cast(
                func.avg(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("avg"),
            cast(
                func.sum(Invocation.remote_duration * JobInstance.multiplier_factor),
                Float,
            ).label("sum"),
            # maxrss
            func.min(Invocation.maxrss).label("min_maxrss"),
            func.max(Invocation.maxrss).label("max_maxrss"),
            cast(func.avg(Invocation.maxrss), Float,).label("avg_maxrss"),
            # avg_cpu
            cast(func.min(Invocation.avg_cpu), Float,).label("min_avg_cpu"),
            cast(func.max(Invocation.avg_cpu), Float,).label("max_avg_cpu"),
            cast(func.avg(Invocation.avg_cpu), Float,).label("avg_avg_cpu"),
        )
        q = q.filter(Invocation.job_instance_id == JobInstance.job_instance_id)
        q = q.filter(Invocation.wf_id.in_(self._wfs))
        q = q.group_by(Invocation.transformation).group_by("type")
        q = q.order_by(Invocation.transformation)

        return q.all()

    #
    # Runtime queries
    # https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
    #

    def _get_date_divisors(self):
        vals = {"month": 2629743, "week": 604800, "day": 86400, "hour": 3600}
        return vals[self._time_filter_mode]

    def _get_host_filter(self):
        if self._host_filter is None:
            return None
        elif isinstance(self._host_filter, str):
            return Host.hostname == self._host_filter
        elif isinstance(self._host_filter, type([])):
            return Host.hostname.in_(self._host_filter)
        else:
            return None

    def _get_xform_filter(self):
        if (
            self._xform_filter["include"] is not None
            and self._xform_filter["exclude"] is not None
        ):
            self.log.error(
                "Can't set both transform include and exclude - reset s.set_transformation_filter()"
            )
            return None
        elif (
            self._xform_filter["include"] is None
            and self._xform_filter["exclude"] is None
        ):
            return None
        elif self._xform_filter["include"] is not None:
            if isinstance(self._xform_filter["include"], str):
                return Invocation.transformation == self._xform_filter["include"]
            elif isinstance(self._xform_filter["include"], type([])):
                return Invocation.transformation.in_(self._xform_filter["include"])
            else:
                return None
        elif self._xform_filter["exclude"] is not None:
            if isinstance(self._xform_filter["exclude"], str):
                return Invocation.transformation != self._xform_filter["exclude"]
            elif isinstance(self._xform_filter["exclude"], type([])):
                return not_(
                    Invocation.transformation.in_(self._xform_filter["exclude"])
                )
            else:
                return None
        else:
            return None

    def get_invocation_by_time(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Invocation.start_time / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            func.count(Invocation.invocation_id).label("count"),
            cast(func.sum(Invocation.remote_duration), Float).label("total_runtime"),
        )
        q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        q = q.filter(Invocation.wf_id == Workflow.wf_id)
        if self._get_xform_filter() is not None:
            q = q.filter(self._get_xform_filter())
        q = q.group_by("date_format").order_by("date_format")

        return q.all()

    def get_jobs_run_by_time(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Jobstate.timestamp / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            func.count(JobInstance.job_instance_id).label("count"),
            cast(func.sum(JobInstance.local_duration), Float).label("total_runtime"),
        )
        q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        q = q.filter(Workflow.wf_id == Job.wf_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(Jobstate.job_instance_id == JobInstance.job_instance_id)
        q = q.filter(Jobstate.state == "EXECUTE")
        q = q.filter(JobInstance.local_duration != None)  # noqa: E711

        if self._get_job_filter() is not None:
            q = q.filter(self._get_job_filter())
        q = q.group_by("date_format").order_by("date_format")

        return q.all()

    def get_invocation_by_time_per_host(self, host=None):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Invocation.start_time / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            Host.hostname.label("host_name"),
            func.count(Invocation.invocation_id).label("count"),
            cast(func.sum(Invocation.remote_duration), Float).label("total_runtime"),
        )
        q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        q = q.filter(Invocation.wf_id == Workflow.wf_id)
        q = q.filter(JobInstance.job_instance_id == Invocation.job_instance_id)
        q = q.filter(JobInstance.host_id == Host.host_id)
        if self._get_host_filter() is not None:
            q = q.filter(self._get_host_filter())
        if self._get_xform_filter() is not None:
            q = q.filter(self._get_xform_filter())
        q = q.group_by("date_format", "host_name").order_by("date_format")

        return q.all()

    def get_jobs_run_by_time_per_host(self):
        """
        https://confluence.pegasus.isi.edu/display/pegasus/Additional+queries
        """
        q = self.session.query(
            (cast(Jobstate.timestamp / self._get_date_divisors(), Integer)).label(
                "date_format"
            ),
            Host.hostname.label("host_name"),
            func.count(JobInstance.job_instance_id).label("count"),
            cast(func.sum(JobInstance.local_duration), Float).label("total_runtime"),
        )
        q = q.filter(Workflow.root_wf_id == self._root_wf_id)
        q = q.filter(Workflow.wf_id == Job.wf_id)
        q = q.filter(Job.job_id == JobInstance.job_id)
        q = q.filter(Jobstate.job_instance_id == JobInstance.job_instance_id)
        q = q.filter(Jobstate.state == "EXECUTE")
        q = q.filter(JobInstance.host_id == Host.host_id)
        if self._get_host_filter() is not None:
            q = q.filter(self._get_host_filter())
        if self._get_job_filter() is not None:
            q = q.filter(self._get_job_filter())
        q = q.group_by("date_format", "host_name").order_by("date_format")

        return q.all()



#!/usr/bin/env python
#
#  Copyright 2017-2021 University Of Southern California
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__author__ = "Rafael Ferreira da Silva"

import logging

from sqlalchemy.exc import *

from Pegasus.db.admin.admin_loader import DBAdminError
from Pegasus.db.admin.versions.base_version import BaseVersion
from Pegasus.db.schema import *

DB_VERSION = 14

log = logging.getLogger(__name__)


class Version(BaseVersion):
    def __init__(self, connection):
        super().__init__(connection)

    def update(self, force=False):
        """."""
        log.debug("Updating to version %s" % DB_VERSION)
        try:
            Trigger.__table__.create(self.db.get_bind(), checkfirst=True)
        except (OperationalError, ProgrammingError):
            pass
        except Exception as e:
            self.db.rollback()
            raise DBAdminError(e)

    def downgrade(self, force=False):
        """."""
        log.debug("Downgrading from version %s" % DB_VERSION)
        # no downgrade is necessary



"""
:mod:`transformation_catalog` exposes an API to serialize and deserialize Pegasus's transformation catalog file.

Basic Usage::

    >>> from Pegasus import transformation_catalog
    >>> transformation_catalog.loads("... ")
    ...

    >>> print(transformation_catalog.dumps( ... ))
    ' ... '

.. moduleauthor:: Ryan Tanaka <tanaka@isi.edu>
.. moduleauthor:: Rajiv Mayani <mayani@isi.edu>
"""

from collections import defaultdict
from io import StringIO
from typing import TextIO

from Pegasus import yaml
from Pegasus.api.errors import PegasusError
from Pegasus.api.site_catalog import OS, Arch
from Pegasus.api.transformation_catalog import (
    Container,
    Transformation,
    TransformationCatalog,
    TransformationSite,
)

__all__ = (
    "load",
    "loads",
    "dump",
    "dumps",
)


def _to_tc(d: dict) -> TransformationCatalog:
    """Convert dict to TransformationCatalog

    :param d: TransformationCatalog represented as a dict
    :type d: dict
    :raises PegasusError: encountered error parsing
    :return: a TransformationCatalog object based on d
    :rtype: TransformationCatalog
    """

    try:
        tc = TransformationCatalog()

        # add transformations
        for tr in d["transformations"]:
            tr_to_add = Transformation(
                tr["name"],
                tr.get("namespace"),
                tr.get("version"),
                checksum=tr.get("checksum"),
            )

            # add transformation sites
            for s in tr["sites"]:
                site_to_add = TransformationSite(
                    s["name"],
                    s["pfn"],
                    True if s["type"] == "stageable" else False,
                    bypass_staging=s.get("bypass"),
                    arch=getattr(Arch, s.get("arch").upper())
                    if s.get("arch")
                    else None,
                    os_type=getattr(OS, s.get("os.type").upper())
                    if s.get("os.type")
                    else None,
                    os_release=s.get("os.release"),
                    os_version=s.get("os.version"),
                    container=s.get("container"),
                )

                # add profiles
                if s.get("profiles"):
                    site_to_add.profiles = defaultdict(dict, s.get("profiles"))

                # add metadata
                if s.get("metadata"):
                    site_to_add.metadata = s.get("metadata")

                # add site to this tr
                tr_to_add.add_sites(site_to_add)

            # add requires
            if tr.get("requires"):
                tr_to_add.requires = set(tr.get("requires"))

            # add profiles
            if tr.get("profiles"):
                tr_to_add.profiles = defaultdict(dict, tr.get("profiles"))

            # add hooks
            if tr.get("hooks"):
                tr_to_add.hooks = defaultdict(list, tr.get("hooks"))

            # add metadata
            if tr.get("metadata"):
                tr_to_add.metadata = tr.get("metadata")

            # add tr to tc
            tc.add_transformations(tr_to_add)

        # add containers
        if "containers" in d:
            for cont in d["containers"]:
                cont_to_add = Container(
                    cont["name"],
                    getattr(Container, cont["type"].upper()),
                    cont["image"],
                    mounts=cont.get("mounts"),
                    image_site=cont.get("image.site"),
                    checksum=cont.get("checksum"),
                    bypass_staging=cont.get("bypass"),
                )

                # add profiles
                if cont.get("profiles"):
                    cont_to_add.profiles = defaultdict(dict, cont.get("profiles"))

                # add cont to tc
                tc.add_containers(cont_to_add)

        return tc

    except KeyError:
        raise PegasusError("error parsing {}".format(d))


def load(fp: TextIO, *args, **kwargs) -> TransformationCatalog:
    """
    Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a TransformationCatalog document) to a :py:class:`~Pegasus.api.transformation_catalog.TransformationCatalog` object.

    :param fp: file like object to load from
    :type fp: TextIO
    :return: deserialized TransformationCatalog object
    :rtype: TransformationCatalog
    """
    return _to_tc(yaml.load(fp))


def loads(s: str, *args, **kwargs) -> TransformationCatalog:
    """
    Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance containing a TransformationCatalog document) to a :py:class:`~Pegasus.api.transformation_catalog.TransformationCatalog` object.

    :param s: string to load from
    :type s: str
    :return: deserialized TransformationCatalog object
    :rtype: TransformationCatalog
    """
    return _to_tc(yaml.load(s))


def dump(
    obj: TransformationCatalog, fp: TextIO, _format="yml", *args, **kwargs
) -> None:
    """
    Serialize ``obj`` as a :py:class:`~Pegasus.api.transformation_catalog.TransformationCatalog` formatted stream to ``fp`` (a ``.write()``-supporting file-like object).

    :param obj: TransformationCatalog to serialize
    :type obj: TransformationCatalog
    :param fp: file like object to serialize to
    :type fp: TextIO
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :rtype: NoReturn
    """
    obj.write(fp, _format=_format)


def dumps(obj: TransformationCatalog, _format="yml", *args, **kwargs) -> str:
    """
    Serialize ``obj`` to a :py:class:`~Pegasus.api.transformation_catalog.TransformationCatalog` formatted ``str``.

    :param obj: TransformationCatalog to serialize
    :type obj: TransformationCatalog
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :return: TransformationCatalog serialized as a string
    :rtype: str
    """
    with StringIO() as s:
        obj.write(s, _format=_format)
        s.seek(0)
        return s.read()



## Copyright (c) 2004-2012, The Regents of the University of California,
## through Lawrence Berkeley National Laboratory
## (subject to receipt of any required approvals from
## the U.S. Dept. of Energy).  All rights reserved.
"""
Software version
"""
NL_VERSION = "4.3.0"
NL_CREATE_DATE = "$Date$"[7:-1].strip()
NL_COPYRIGHT = """Copyright (c) 2004-2012, The Regents of the University of California, through
Lawrence Berkeley National Laboratory (subject to receipt of any required
approvals from the U.S. Dept. of Energy).  All rights reserved."""



"""
:mod:`replica_catalog` exposes an API to serialize and deserialize Pegasus's replica catalog file.

Basic Usage::

    >>> from Pegasus import replica_catalog
    >>> replica_catalog.loads("... ")
    ...

    >>> print(replica_catalog.dumps( ... ))
    ' ... '

.. moduleauthor:: Ryan Tanaka <tanaka@isi.edu>
.. moduleauthor:: Rajiv Mayani <mayani@isi.edu>
"""

from io import StringIO
from typing import TextIO

from Pegasus import yaml
from Pegasus.api.errors import PegasusError
from Pegasus.api.replica_catalog import _PFN, ReplicaCatalog

__all__ = (
    "load",
    "loads",
    "dump",
    "dumps",
)


def _to_rc(d: dict) -> ReplicaCatalog:
    """Convert dict to ReplicaCatalog

    :param d: ReplicaCatalog represented as a dict
    :type d: dict
    :raises PegasusError: encountered error parsing
    :return: a ReplicaCatalog object based on d
    :rtype: ReplicaCatalog
    """
    rc = ReplicaCatalog()

    try:
        for r in d["replicas"]:
            lfn = r["lfn"]
            pfns = {_PFN(i["site"], i["pfn"]) for i in r["pfns"]}

            checksum = r.get("checksum") if r.get("checksum") else {}
            metadata = r.get("metadata") if r.get("metadata") else {}

            regex = r.get("regex")
            for pfn in pfns:
                if regex:
                    rc.add_regex_replica(pfn.site, lfn, pfn.pfn, metadata=metadata)
                else:
                    rc.add_replica(
                        pfn.site, lfn, pfn.pfn, metadata=metadata, checksum=checksum
                    )

    except KeyError:
        raise PegasusError("error parsing {}".format(d))

    return rc


def load(fp: TextIO, *args, **kwargs) -> ReplicaCatalog:
    """
    Deserialize ``fp`` (a ``.read()``-supporting file-like object containing a ReplicaCatalog document) to a :py:class:`~Pegasus.api.replica_catalog.ReplicaCatalog` object.

    :param fp: file like object to load from
    :type fp: TextIO
    :return: deserialized ReplicaCatalog object
    :rtype: ReplicaCatalog
    """
    return _to_rc(yaml.load(fp))


def loads(s: str, *args, **kwargs) -> ReplicaCatalog:
    """
    Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance containing a ReplicaCatalog document) to a :py:class:`~Pegasus.api.replica_catalog.ReplicaCatalog` object.

    :param s: string to load from
    :type s: str
    :return: deserialized ReplicaCatalog object
    :rtype: ReplicaCatalog
    """
    return _to_rc(yaml.load(s))


def dump(obj: ReplicaCatalog, fp: TextIO, _format="yml", *args, **kwargs) -> None:
    """
    Serialize ``obj`` as a :py:class:`~Pegasus.api.replica_catalog.ReplicaCatalog` formatted stream to ``fp`` (a ``.write()``-supporting file-like object).

    :param obj: ReplicaCatalog to serialize
    :type obj: ReplicaCatalog
    :param fp: file like object to serialize to
    :type fp: TextIO
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :rtype: NoReturn
    """
    obj.write(fp, _format=_format)


def dumps(obj: ReplicaCatalog, _format="yml", *args, **kwargs) -> str:
    """
    Serialize ``obj`` to a :py:class:`~Pegasus.api.replica_catalog.ReplicaCatalog` formatted ``str``.

    :param obj: ReplicaCatalog to serialize
    :type obj: ReplicaCatalog
    :param _format: format to write to if fp does not have an extension; can be one of ["yml" | "yaml" | "json"], defaults to "yml"
    :type _format: str
    :return: ReplicaCatalog serialized as a string
    :rtype: str
    """
    with StringIO() as s:
        obj.write(s, _format=_format)
        s.seek(0)
        return s.read()
